/*
 * Copyright 2012, 2017 INFN
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they
 * will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the
 * Licence.
 * You may obtain a copy of the Licence at:
 *
 * https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in
 * writing, software distributed under the Licence is
 * distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied.
 * See the Licence for the specific language governing
 * permissions and limitations under the Licence.
 */

#include <chaos/common/configuration/GlobalConfiguration.h>
#include <chaos/common/event/channel/InstrumentEventChannel.h>
#include <chaos/common/global.h>
#include <chaos/common/healt_system/HealtManager.h>
#include <chaos/common/property/property.h>
#include <chaos/common/utility/UUIDUtil.h>
#include "../command_manager/CommandManager.h"
#include "AbstractControlUnit.h"
#include "../data_manager/DataManager.h"
#include "../driver_manager/DriverManager.h"
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/uuid/uuid.hpp>             // uuid class
#include <boost/uuid/uuid_generators.hpp>  // generators
#include <boost/uuid/uuid_io.hpp>          // streaming operators etc.
#include <boost/filesystem.hpp>

#include "../windowsCompliant.h"
using namespace boost::uuids;
using namespace chaos::common::data;
using namespace chaos::common::data::structured;
using namespace chaos::common::alarm;
using namespace chaos::common::utility;
using namespace chaos::common::property;
using namespace chaos::common::exception;
using namespace chaos::common::data::cache;
using namespace chaos::common::healt_system;
using namespace chaos::common::metadata_logging;

using namespace chaos::cu::data_manager;
using namespace chaos::cu::control_manager;
using namespace chaos::cu::driver_manager;
using namespace chaos::cu::driver_manager::driver;

#define DS_UPDATE_ANYWAY_DEF 60000
#define CONTROL_UNIT_LOG_MAX_MS_DEF 10000
#define DBG DBG_LOG(AbstractControlUnit)
#define ERR ERR_LOG(AbstractControlUnit)

#define ACULAPP_ LAPP_ << "[Control Unit:" << control_unit_instance << "-" << control_unit_id << "] -"
#define ACULNOTE_ LNOTE_ << "[Control Unit:" << control_unit_instance << "-" << control_unit_id << "] -"
#define ACULWRN_ LWRN_ << "[Control Unit:" << control_unit_instance << "-" << control_unit_id << "] -"
#define ACULDBG_ DBG << "[Control Unit:" << control_unit_instance << "-" << control_unit_id << "] -" << __FUNCTION__ << "-"
#define ACULERR_ ERR << "[Control Unit:" << control_unit_instance << "-" << control_unit_id << "](-" << __FUNCTION__ << "-" << __LINE__ << ") - "

#define S(x) #x
#define S_(x) S(x)
#define S__LINE__ S_(__LINE__)

#define CHEK_IF_NEED_TO_THROW(flag, code)                                                                                          \
  try {                                                                                                                            \
    code                                                                                                                           \
  } catch (chaos::CException & ex) {                                                                                               \
    ACULERR_ << "CHAOS Exception on " << DatasetDB::getDeviceID() << ":\n"                                                         \
             << ex.what();                                                                                                         \
    if (flag) throw chaos::common::exception::MetadataLoggingCException(getCUID(), ex.errorCode, ex.errorMessage, ex.errorDomain); \
  } catch (...) {                                                                                                                  \
    ACULERR_ << "Unknown exception on" << DatasetDB::getDeviceID();                                                                \
    if (flag) throw chaos::common::exception::MetadataLoggingCException(getCUID(), -1000, S__LINE__, __PRETTY_FUNCTION__);         \
  }

#define GET_CAT_OR_EXIT(t, rv)              \
  if (map_variable_catalog.count(t) == 0) { \
    return rv;                              \
  }                                         \
  AlarmCatalog& catalog = map_variable_catalog[t];

#pragma mark StorageBurst
StorageBurst::StorageBurst(DatasetBurstShrdPtr _dataset_burst)
    : dataset_burst(_dataset_burst), _remaining(0) {}

StorageBurst::~StorageBurst() {}

#pragma mark PushStorageBurst
PushStorageBurst::PushStorageBurst(DatasetBurstShrdPtr _dataset_burst)
    : StorageBurst(MOVE(_dataset_burst)), current_pushes(0) {}

PushStorageBurst::~PushStorageBurst() {}
#ifndef _WIN32
bool PushStorageBurst::active(int64_t data __attribute__((unused))) {
#else
bool PushStorageBurst::active(int64_t data) {
#endif
  _remaining = StorageBurst::dataset_burst->value.asUInt32() - current_pushes - 1;
  return ++current_pushes < StorageBurst::dataset_burst->value.asUInt32();
}

#pragma mark MSecStorageBurst
MSecStorageBurst::MSecStorageBurst(DatasetBurstShrdPtr _dataset_burst)
    : StorageBurst(MOVE(_dataset_burst)), timeout_msec(TimingUtil::getTimestampWithDelay(StorageBurst::dataset_burst->value.asInt32(), true)) {}

MSecStorageBurst::~MSecStorageBurst() {}

bool MSecStorageBurst::active(int64_t now) {
  _remaining = timeout_msec - now;
  return timeout_msec > now;
}

#pragma mark AbstractControlUnit
//! Contructor with type and id
std::string AbstractControlUnit::base_data_path;

AbstractControlUnit::AbstractControlUnit(const std::string& _control_unit_type,
                                         const std::string& _control_unit_id,
                                         const std::string& _control_unit_param)
    : DatasetDB(GlobalConfiguration::getInstance()->getOption<bool>(CU_OPT_IN_MEMORY_DATABASE)), control_key("none"),
#ifdef CONTROL_UNIT_INSTANCE_RANDOM
    control_unit_instance(UUIDUtil::generateUUIDLite())
    ,
#else
    control_unit_instance(_control_unit_id)
    ,
#endif
    control_unit_type(_control_unit_type)
    , control_unit_id(_control_unit_id)
    , control_unit_param(_control_unit_param)
    , run_id(0)
    , standard_logging_channel()
    , alarm_logging_channel()
    , push_dataset_counter(0)
    , aligned_ex_done(false)
    , push_dataset_size(0)
    , push_tot_size(0)
    , last_push_rate_grap_ts(0)
    , attribute_value_shared_cache()
    , attribute_shared_cache_wrapper()
    , use_custom_high_resolution_timestamp(false)
    , timestamp_acq_cached_value()
    , timestamp_hw_acq_cached_value()
    , thread_schedule_daly_cached_value()
    , ds_update_anyway(DS_UPDATE_ANYWAY_DEF)
    , log_maxupdate_ms(CONTROL_UNIT_LOG_MAX_MS_DEF)
    , last_push(0)
    , key_data_storage()
    , busy(false),bypass(false) {
  _initPropertyGroup();
  //!try to decode parameter string has json document
  is_control_unit_json_param = CDataWrapper::isJSON(control_unit_param);
  //initialize check list
if(base_data_path==""){
      base_data_path=GlobalConfiguration::getInstance()->getOption< std::string>(InitOption::OPT_DATA_DIR);

  }
  _initChecklist();
  
}

//! Contructor with driver
AbstractControlUnit::AbstractControlUnit(const std::string&           _control_unit_type,
                                         const std::string&           _control_unit_id,
                                         const std::string&           _control_unit_param,
                                         const ControlUnitDriverList& _control_unit_drivers)
    : DatasetDB(GlobalConfiguration::getInstance()->getOption<bool>(CU_OPT_IN_MEMORY_DATABASE)), control_key("none"),
#ifdef CONTROL_UNIT_INSTANCE_RANDOM
    control_unit_instance(UUIDUtil::generateUUIDLite())
    ,
#else
    control_unit_instance(_control_unit_id)
    ,
#endif
    control_unit_type(_control_unit_type)
    , control_unit_class("AbstractControlUnit")
    , control_unit_id(_control_unit_id)
    , control_unit_param(_control_unit_param)
    , run_id(0)
    , hasstopped(true)
    , standard_logging_channel()
    , alarm_logging_channel()
    , push_dataset_counter(0)
    , aligned_ex_done(false)
    , push_errors(0)
    , packet_lost(0)
    , last_push_rate_grap_ts(0)
    , attribute_value_shared_cache()
    , attribute_shared_cache_wrapper()
    , use_custom_high_resolution_timestamp(false)
    , timestamp_acq_cached_value()
    , timestamp_hw_acq_cached_value()
    , thread_schedule_daly_cached_value()
    , key_data_storage(), busy(false),bypass(false) {
  _initPropertyGroup();
  //!try to decode parameter string has json document
  is_control_unit_json_param = CDataWrapper::isJSON(control_unit_param);
  if(base_data_path==""){
      base_data_path=GlobalConfiguration::getInstance()->getOption< std::string>(InitOption::OPT_DATA_DIR);

  }
  //copy array
  for (int idx = 0; idx < _control_unit_drivers.size(); idx++) {
    control_unit_drivers.push_back(_control_unit_drivers[idx]);
  }

  //initialize check list
  _initChecklist();
}

void AbstractControlUnit::_initDrivers() {
  ACULDBG_ << "Initializating Driver Accessors";
  //at this point and before the unit implementation init i need to get
  //the infromation about the needed drivers
  std::vector<DrvRequestInfo> unit_needed_drivers;

  //got the needded driver definition
  unitDefineDriver(unit_needed_drivers);

  accessor_instances.clear();
  for (int idx = 0;
       idx != unit_needed_drivers.size();
       idx++) {
    driver_manager::driver::DriverAccessor* accessor_instance = driver_manager::DriverManager::getInstance()->getNewAccessorForDriverInstance(unit_needed_drivers[idx], control_unit_id);
    accessor_instances.push_back(accessor_instance);
  }
}
void AbstractControlUnit::setCUClass(const std::string& cl) {
  control_unit_class = cl;
}

void AbstractControlUnit::_initChecklist() {
  //init checklists
  check_list_sub_service.addCheckList("_init");
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_CALL_INIT_STATE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_INIT_SHARED_CACHE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_COMPLETE_OUTPUT_ATTRIBUTE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_COMPLETE_INPUT_ATTRIBUTE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_INIT_SYSTEM_CACHE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_CALL_UNIT_DEFINE_ATTRIBUTE);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_CREATE_FAST_ACCESS_CASCHE_VECTOR);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_CALL_UNIT_INIT);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_UPDATE_CONFIGURATION);
  check_list_sub_service.getSharedCheckList("_init")->addElement(INIT_RPC_PHASE_PUSH_DATASET);

  check_list_sub_service.addCheckList("init");
  check_list_sub_service.getSharedCheckList("init")->addElement(INIT_SM_PHASE_INIT_DB);
  check_list_sub_service.getSharedCheckList("init")->addElement(INIT_SM_PHASE_CREATE_DATA_STORAGE);

  check_list_sub_service.addCheckList("_start");
  check_list_sub_service.getSharedCheckList("_start")->addElement(START_RPC_PHASE_UNIT);
  check_list_sub_service.getSharedCheckList("_start")->addElement(START_RPC_PHASE_IMPLEMENTATION);

  check_list_sub_service.addCheckList("start");
  check_list_sub_service.getSharedCheckList("start")->addElement(START_SM_PHASE_STAT_TIMER);
}

void AbstractControlUnit::_initPropertyGroup() {
  PropertyGroup& pg_abstract_cu = addGroup(chaos::ControlUnitPropertyKey::P_GROUP_NAME);
  pg_abstract_cu.addProperty(ControlUnitDatapackSystemKey::BYPASS_STATE, "Put control unit in bypass state", DataType::TYPE_BOOLEAN, 0, CDataVariant((bool)false));
  pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_STORAGE_TYPE, "Set the control unit storage type", DataType::TYPE_INT32, 0, CDataVariant((int32_t)0));
  pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_STORAGE_LIVE_TIME, "Set the interval time between two live packets", DataType::TYPE_INT64, 0, CDataVariant((int64_t)0));
  pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_STORAGE_HISTORY_TIME, "Set the interval time between two history packets", DataType::TYPE_INT64, 0, CDataVariant((int64_t)0));
  pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_STORAGE_LOG_TIME, "Set the interval time between two log packets", DataType::TYPE_INT64, 0, CDataVariant((int64_t)0));

  pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_UPDATE_ANYWAY, "Update the dataset anyway (ms)", DataType::TYPE_INT32, 0, CDataVariant((int32_t)DS_UPDATE_ANYWAY_DEF));
  pg_abstract_cu.addProperty(ControlUnitDatapackSystemKey::CU_LOG_MAX_MS, "Maximum log rate (ms) 0 always", DataType::TYPE_INT32, 0, CDataVariant((int32_t)CONTROL_UNIT_LOG_MAX_MS_DEF));
  busy=false;bypass=false;
  log_maxupdate_ms = CONTROL_UNIT_LOG_MAX_MS_DEF;
  ds_update_anyway = DS_UPDATE_ANYWAY_DEF;
  //    CDWUniquePtr burst_type_desc(new CDataWrapper());
  //    burst_type_desc->addInt32Value(DataServiceNodeDefinitionKey::DS_HISTORY_BURST_TYPE, DataServiceNodeDefinitionType::DSStorageBurstTypeUndefined);
  //    pg_abstract_cu.addProperty(DataServiceNodeDefinitionKey::DS_HISTORY_BURST, "Specify if the restore operation need to be done as real operation or not", DataType::TYPE_JSON,0, CDataVariant(burst_type_desc.release()));

  pg_abstract_cu.addProperty(ControlUnitDatapackSystemKey::THREAD_SCHEDULE_DELAY, "Set the control unit step repeat time in microseconds", DataType::TYPE_INT64, 0, CDataVariant((int64_t)1000000));  //set to one seconds
  pg_abstract_cu.addProperty(ControlUnitPropertyKey::INIT_RESTORE_OPTION, "Specify the restore type operatio to do durint initialization phase", DataType::TYPE_INT32, 0, CDataVariant((int32_t)0));
  pg_abstract_cu.addProperty(ControlUnitPropertyKey::INIT_RESTORE_APPLY, "Specify if the restore operation need to be done as real operation or not", DataType::TYPE_BOOLEAN, 0, CDataVariant((bool)false));

  PropertyCollector::setPropertyValueChangeFunction(ChaosBind(&AbstractControlUnit::propertyChangeHandler, this, ChaosBindPlaceholder(_1), ChaosBindPlaceholder(_2), ChaosBindPlaceholder(_3)));
  PropertyCollector::setPropertyValueUpdatedFunction(ChaosBind(&AbstractControlUnit::propertyUpdatedHandler, this, ChaosBindPlaceholder(_1), ChaosBindPlaceholder(_2), ChaosBindPlaceholder(_3), ChaosBindPlaceholder(_4)));
}

/*!
     Destructor a new CU with an identifier
     */
AbstractControlUnit::~AbstractControlUnit() {
  //clear the accessor of the driver
  for (int idx = 0;
       idx != accessor_instances.size();
       idx++) {
    driver_manager::DriverManager::getInstance()->releaseAccessor(accessor_instances[idx]);
  }
  //clear the vector
  accessor_instances.clear();
}

/*
     return the CU name
     */
const std::string& AbstractControlUnit::getCUInstance() {
  return control_unit_instance;
};

const std::string& AbstractControlUnit::getCUID() {
  return control_unit_id;
}

const std::string& AbstractControlUnit::getCUParam() {
  return control_unit_param;
}

int AbstractControlUnit::getCUParam(chaos::common::data::CDataWrapper& p) {
  try {
    if (control_unit_param != "") {
      p.reset();
      p.setSerializedJsonData(control_unit_param.c_str());
      return 0;
    }
  } catch (...) {
  }
  return -1;
}

const bool AbstractControlUnit::isCUParamInJson() {
  return is_control_unit_json_param;
}

const std::string& AbstractControlUnit::getCUType() {
  return control_unit_type;
}

void AbstractControlUnit::setAlarm(const std::string& name, int level) {
  common::alarm::MultiSeverityAlarmLevel lvl = MultiSeverityAlarmLevelClear;
  if (level == 1) {
    lvl = MultiSeverityAlarmLevelWarning;
  } else if (level >= 2) {
    lvl = MultiSeverityAlarmLevelHigh;
  }
  setStateVariableSeverity(StateVariableTypeAlarmCU, name, lvl);
}

/*
     fill the CDataWrapper with AbstractCU system configuration, this method
     is called after getStartConfiguration directly by sandbox. in this method
     are defined the action for the input element of the dataset
     */
void AbstractControlUnit::_defineActionAndDataset(CDataWrapper& setup_configuration) {
  vector<std::string> tempStringVector;

  if (control_unit_id.size()) {
    setDeviceID(control_unit_id);
  }

  //add the CU isntance, this can be redefinide by user in the unitDefineActionAndDataset method
  //for let the CU have the same instance at every run
  setup_configuration.addStringValue(NodeDefinitionKey::NODE_RPC_DOMAIN, control_unit_instance);

  //if we have a control key we attach it
  setup_configuration.addStringValue("mds_control_key", control_key);

  //add the control unit type with semantonc type::subtype
  setup_configuration.addStringValue(NodeDefinitionKey::NODE_TYPE, NodeType::NODE_TYPE_CONTROL_UNIT);
  setup_configuration.addStringValue(NodeDefinitionKey::NODE_SUB_TYPE, control_unit_type);
  //check if as been setuped a file for configuration
  //LCU_ << "Check if as been setup a json file path to configura CU:" << CU_IDENTIFIER_C_STREAM;
  //loadCDataWrapperForJsonFile(setup_configuration);

  //first call the setup abstract method used by the implementing CU to define action, dataset and other
  //usefull value
  unitDefineActionAndDataset();

  addStateVariable(StateVariableTypeAlarmCU, "packet_send_error", "Notify when a CU cannot push datasset");
  addStateVariable(StateVariableTypeAlarmCU, "packet_lost", "Notify when a packet is definitively lost");

  //call method to dinamically add other things to the dataset
  _completeDatasetAttribute();

  //for now we need only to add custom action for expose to rpc
  //input element of the dataset
  AbstActionDescShrPtr
      action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                            &AbstractControlUnit::_setDatasetAttribute,
                                                                            ControlUnitNodeDomainAndActionRPC::CONTROL_UNIT_APPLY_INPUT_DATASET_ATTRIBUTE_CHANGE_SET,
                                                                            "method for setting the input element for the dataset");

  //expose updateConfiguration Methdo to rpc
  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::updateConfiguration,
                                                                        NodeDomainAndActionRPC::ACTION_UPDATE_PROPERTY,
                                                                        "Update control unit property");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_init,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_INIT,
                                                                        "Perform the control unit initialization");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_deinit,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_DEINIT,
                                                                        "Perform the control unit deinitialization");
  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_start,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_START,
                                                                        "Start the control unit scheduling");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_stop,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_STOP,
                                                                        "Stop the control unit scheduling");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_recover,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_RECOVER,
                                                                        "Recovery a recoverable state, going to the last state");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_unitRestoreToSnapshot,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_RESTORE,
                                                                        "Restore control unit to a snapshot tag");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_unitPerformCalibration,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_CALIBRATION,
                                                                        "Perform Calibration");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_getState,
                                                                        NodeDomainAndActionRPC::ACTION_NODE_GET_STATE,
                                                                        "Get the state of the running control unit");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_getInfo,
                                                                        NodeDomainAndActionRPC::ACTION_CU_GET_INFO,
                                                                        "Get the information about running control unit");
  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_submitStorageBurst,
                                                                        ControlUnitNodeDomainAndActionRPC::ACTION_STORAGE_BURST,
                                                                        "Execute a storage burst on control unit");

  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::setProperty, NodeDomainAndActionRPC::ACTION_SET_PROPERTIES, "method for set properties of a CU");
  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::getProperty, NodeDomainAndActionRPC::ACTION_GET_PROPERTIES, "method for get properties of a CU");

  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::_setDriverProperties, ControlUnitNodeDomainAndActionRPC::CONTROL_UNIT_DRV_SET_PROPERTIES, "method for set properties of a driver");
  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::_getDriverProperties, ControlUnitNodeDomainAndActionRPC::CONTROL_UNIT_DRV_GET_PROPERTIES, "method for get properties of a driver");
  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::clearAlarm, NodeDomainAndActionRPC::ACTION_NODE_CLRALRM, "method for clear alarm");

  addActionDescritionInstance<AbstractControlUnit>(this, &AbstractControlUnit::setAlarm, ControlUnitNodeDomainAndActionRPC::CONTROL_UNIT_SET_ALARM, "method for set alarms and  masks");

  action_description->addParam(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_HISTORY_BURST_TAG, DataType::TYPE_STRING, "Tag associated to the stored data during burst");
  action_description->addParam(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_HISTORY_BURST_TYPE, DataType::TYPE_INT32, "The type of burst");
  action_description->addParam(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_HISTORY_BURST_VALUE, DataType::TYPE_UNDEFINED, "The value of the burst is defined by the type");

  action_description = addActionDescritionInstance<AbstractControlUnit>(this,
                                                                        &AbstractControlUnit::_datasetTagManagement,
                                                                        ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT,
                                                                        "Execute a storage burst on control unit");
  action_description->addParam(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_ADD_LIST, DataType::TYPE_ACCESS_ARRAY, "List of tag to be added to the control unit dataset");
  action_description->addParam(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST, DataType::TYPE_ACCESS_ARRAY, "List of tag to be removed to the control unit dataset");
  //grab dataset description
  DatasetDB::fillDataWrapperWithDataSetDescription(setup_configuration);

  //    //get action description
  //    getActionDescrionsInDataWrapper(setup_configuration);

  //add property description
  PropertyCollector::fillDescription("property", setup_configuration);
}
chaos::common::data::CDWUniquePtr AbstractControlUnit::getProperty(chaos::common::data::CDWUniquePtr data) {
  bool sync=true;
   chaos::common::data::CDWUniquePtr ret ;
  if(data.get()&&data->hasKey("sync")){
    sync=data->getBoolValue("sync");
  }
  if(data.get()&&data->hasKey("name")){
    std::string name=data->getStringValue("name");
    ret=syncRead(name);
    ACULDBG_ << "get CU property :" <<name<<" ="<< ((ret.get()) ? ret->getJSONString() : "");

    return ret;
  }
  ret= getProperties(sync);
  ACULDBG_ << "get CU properties:" << ((ret.get()) ? ret->getJSONString() : "");
  return ret;
}
chaos::common::data::CDWUniquePtr AbstractControlUnit::setProperty(chaos::common::data::CDWUniquePtr data) {
  if (data.get()) {
    setProperties(*data.get(), true);
  }
  return getProperties();
}
 
int AbstractControlUnit::setStateMask(const std::string& name,bool maskunmask){
  AlarmCatalog&     catalogcu  = map_variable_catalog[StateVariableTypeAlarmCU];
  AlarmCatalog&     catalogdev = map_variable_catalog[StateVariableTypeAlarmDEV];
  AlarmDescription* alarmc     = catalogcu.getAlarmByName(name);
  AlarmDescription* alarmv     = catalogdev.getAlarmByName(name);
  int mask=0xFF;
  if(maskunmask){
    mask=0;
  }
  int found=0;
  if (alarmc != NULL) {
    ACULDBG_ << "Set CU alarm \"" << name << "\" mask to:" << mask;

    alarmc->setMask(mask);
    found++;
  }
  if (alarmv != NULL) {
    ACULDBG_ << "Set DEV alarm \"" << name << "\" mask to:" << mask;

    alarmv->setMask(mask);
    found++;

  }
  return found;
}
 
int AbstractControlUnit::setStateAllMask(bool maskunmask){
  AlarmCatalog&     catalogcu  = map_variable_catalog[StateVariableTypeAlarmCU];
  AlarmCatalog&     catalogdev = map_variable_catalog[StateVariableTypeAlarmDEV];
  int mask=0xFF;
  if(maskunmask){
    mask=0;
  }
  catalogcu.setAllAlarmMask(mask);
   catalogdev.setAllAlarmMask(mask);
  return (catalogcu.size()+catalogdev.size());
}
void AbstractControlUnit::setAlarmMask(const std::string& name, uint32_t mask) {
  AlarmCatalog&     catalogcu  = map_variable_catalog[StateVariableTypeAlarmCU];
  AlarmCatalog&     catalogdev = map_variable_catalog[StateVariableTypeAlarmDEV];
  AlarmDescription* alarmc     = catalogcu.getAlarmByName(name);
  AlarmDescription* alarmv     = catalogdev.getAlarmByName(name);
  if (alarmc != NULL) {
    ACULDBG_ << "Set CU alarm \"" << name << "\" mask to:" << mask;

    alarmc->setMask(mask);
  }
  if (alarmv != NULL) {
    ACULDBG_ << "Set DEV alarm \"" << name << "\" mask to:" << mask;

    alarmv->setMask(mask);
  }
}
  int AbstractControlUnit::clearData(const std::string& keyname){
     ChaosSharedPtr<SharedCacheLockDomain> w_lock = attribute_value_shared_cache->getLockOnDomain((SharedCacheDomain)DOMAIN_CUSTOM, true);
    w_lock->lock();

    std::string fname = control_unit_id;
    replace(fname.begin(), fname.end(), '/', '_');
    std::stringstream ss;
    ss << base_data_path<<"/"<<fname;
    boost::filesystem::path p(ss.str());
    if ((boost::filesystem::exists(p))) {
      boost::filesystem::remove(p);
    }
    /* no still possible
    if(getAttributeCache()->exist(DOMAIN_CUSTOM,keyname)){
        getAttributeCache()->removeCustomAttribute(keyname);
        fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                                  cache_custom_attribute_vector);
        getAttributeCache()->setCustomDomainAsChanged();
      pushCustomDataset();
      
    }
    */

    w_lock->unlock();
    return 0;

  }

  int AbstractControlUnit::saveData(const std::string& keyname,const chaos::common::data::CDataWrapper& d){
    ChaosSharedPtr<SharedCacheLockDomain> w_lock = attribute_value_shared_cache->getLockOnDomain((SharedCacheDomain)DOMAIN_CUSTOM, true);
    w_lock->lock();

    std::string fname = control_unit_id;
    replace(fname.begin(), fname.end(), '/', '_');
    std::stringstream ss;
    ss << base_data_path<<"/"<<fname;
    boost::filesystem::path p(ss.str());
    if ((boost::filesystem::exists(p) == false)) {
          try {
            if ((boost::filesystem::create_directories(p) == false) && ((boost::filesystem::exists(p) == false))) {
              ACULERR_ << "cannot create directory:" << p ;
              return -3;
            } else {
              ACULDBG_ << " CREATED DIR:" << p;
            }
          } catch (boost::filesystem::filesystem_error& e) {
            ACULERR_ << " Exception creating directory:" << e.what();
            return -1;
          }
          //
      }
    ss<<"/"+keyname;
    std::ofstream fs;
    fs.open(ss.str().c_str());
    if(!fs.is_open()){
        ACULERR_ << " cannot create " << ss.str();
        return -5;
    }
    fs<<d.getJSONString();

    fs.close();
    if(!getAttributeCache()->exist(DOMAIN_CUSTOM,keyname)){
      getAttributeCache()->addCustomAttribute(keyname, d);
      ACULDBG_ << "CREATE custom attribute:'"<<keyname<<"'";

    } 
    getAttributeCache()->setCustomAttributeValue(keyname, d);

    
    ACULDBG_ << keyname<<" wrote " <<ss.str()<<" size:"<<d.getJSONString().size() << "in "<<ss.str().c_str();

    fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                                  cache_custom_attribute_vector);
    getAttributeCache()->setCustomDomainAsChanged();
    pushCustomDataset();
    return 0;
  }
  
  chaos::common::data::CDWUniquePtr AbstractControlUnit::loadData(const std::string& keyname){
    ChaosSharedPtr<SharedCacheLockDomain> w_lock = attribute_value_shared_cache->getLockOnDomain((SharedCacheDomain)DOMAIN_CUSTOM, true);
    w_lock->lock();
    chaos::common::data::CDWUniquePtr ret;
    std::string fname = control_unit_id;
    replace(fname.begin(), fname.end(), '/', '_');
    std::stringstream ss;
    ss <<base_data_path <<"/"<<fname<<"/"+keyname;
    std::ifstream fs;
    fs.open(ss.str().c_str());
    if(fs.is_open()){
      std::stringstream buffer;
      buffer << fs.rdbuf();
      chaos::common::data::CDataWrapper*w=new chaos::common::data::CDataWrapper();
      if(w){
        try {
          w->setSerializedJsonData(buffer.str().c_str());
          ret.reset(w);
          if(!getAttributeCache()->exist(DOMAIN_CUSTOM,keyname)){
              ACULDBG_ << "CREATE custom attribute:'"<<keyname<<"' "<<w->getJSONString();

            getAttributeCache()->addCustomAttribute(keyname, *w);
          } 
            getAttributeCache()->setCustomAttributeValue(keyname, *w);

            getAttributeCache()->setCustomDomainAsChanged();

           fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                                  cache_custom_attribute_vector);
           pushCustomDataset();
      
          ACULDBG_ << keyname<<" read: " <<ss.str()<<" size:"<<buffer.str().size();
        } catch(...){
            ACULERR_ << " parsing JSON " << buffer.str();

        }
      }

    }
    return ret;
  }
chaos::common::data::CDWUniquePtr AbstractControlUnit::clearAlarm(chaos::common::data::CDWUniquePtr data) {
  if (data.get()) {
    if (!data->hasKey("value")) {
      data->addInt32Value("value", 0);
    }
     if (!data->hasKey("all")) {
      data->addBoolValue("all",true);
    }
    return setAlarm(MOVE(data));
  }
  return data;
}
chaos::common::data::CDWUniquePtr AbstractControlUnit::setAlarm(chaos::common::data::CDWUniquePtr data) {
    ACULDBG_ << "Set Alarm "<< (data.get()?data->getJSONString():" NO DATA");

  if (data.get()) {
    bool              mod           = false;
    AlarmCatalog&     catalogcu     = map_variable_catalog[StateVariableTypeAlarmCU];
    AlarmCatalog&     catalogdev    = map_variable_catalog[StateVariableTypeAlarmDEV];
    AlarmCatalog&     catalog       = catalogcu;
    StateVariableType variable_type = StateVariableTypeAlarmCU;

    if (data->hasKey("name")) {
      AlarmDescription* alarm = catalogcu.getAlarmByName(data->getStringValue("name"));
      if (alarm == NULL) {
        alarm         = catalogdev.getAlarmByName(data->getStringValue("name"));
        variable_type = StateVariableTypeAlarmDEV;
        catalog       = catalogdev;
      }
      if (alarm == NULL) {
        ACULERR_ << "Alarm \"" << data->getStringValue("name") << " NOT FOUND";
        return chaos::common::data::CDWUniquePtr();
      }
      if (data->hasKey("mask") && data->isInt32Value("mask")) {
        uint32_t m = data->getInt32Value("mask");
        //  ACULDBG_ << "Set mask "<< data->getStringValue("name")<<" to :" << m;

        alarm->setMask(m);
        mod = true;
      }
      if (data->hasKey("value") && data->isInt32Value("value")) {
        alarm->setCurrentSeverity(data->getInt32Value("value"));
        mod = true;
      }

    } else if (data->hasKey("all")) {
      AlarmCatalog& catalogcu  = map_variable_catalog[StateVariableTypeAlarmCU];
      AlarmCatalog& catalogdev = map_variable_catalog[StateVariableTypeAlarmDEV];
      if (data->hasKey("value")) {
        int32_t val = data->getInt32Value("value");
        ACULDBG_ << "Set value of all Alarms to :" << val;
        mod = true;
        catalogcu.setAllAlarmSeverity(val);
        catalogdev.setAllAlarmSeverity(val);
      }
      if (data->hasKey("mask") && data->isInt32Value("mask")) {
        uint32_t val = data->getInt32Value("mask");
        mod=true;
        ACULDBG_ << "Set mask of all Alarms to :" << val;
        catalogcu.setAllAlarmMask(val);
        catalogdev.setAllAlarmMask(val);

        
      }
    }
    if (mod) {
      AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
      output_cache.getValueSettingByName(stateVariableEnumToName(StateVariableTypeAlarmCU))->setValue(CDataVariant(catalogcu.maxLevel()));
      output_cache.getValueSettingByName(stateVariableEnumToName(StateVariableTypeAlarmDEV))->setValue(CDataVariant(catalogdev.maxLevel()));
      //output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.maxLevel()));
      pushDevAlarmDataset();
      pushCUAlarmDataset();
      pushSystemDataset();
      HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_ALARM_LEVEL,
                                                  std::max(catalogcu.maxLevel(), catalogdev.maxLevel()));

      HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_ALARM_MASKED,
                                                  catalogcu.countMask()+ catalogdev.countMask());

    }
  }

  return chaos::common::data::CDWUniquePtr();
}

chaos::common::data::CDWUniquePtr AbstractControlUnit::_setDriverProperties(chaos::common::data::CDWUniquePtr data) {
  if (data.get()) {
    for (VInstantitedDriverIterator it  = accessor_instances.begin(),
                                    end = accessor_instances.end();
         it != end;
         it++) {
      if (data->hasKey("_id_")) {
        if ((*it)->getDriverName() == data->getStringValue("_id_")) {
          ACULDBG_ << "set driver " << (*it)->getDriverName() << " property:" << data->getJSONString();
          return (*it)->setDrvProperties(data);
        }
      } else {
        ACULDBG_ << "set driver " << (*it)->getDriverName() << " property:" << data->getJSONString();

        return (*it)->setDrvProperties(data);
      }
    }
  }
  return chaos::common::data::CDWUniquePtr();
}
chaos::common::data::CDWUniquePtr AbstractControlUnit::_getDriverProperties(chaos::common::data::CDWUniquePtr data) {
  chaos::common::data::CDWUniquePtr ret;

  for (VInstantitedDriverIterator it  = accessor_instances.begin(),
                                  end = accessor_instances.end();
       it != end;
       it++) {
    if (data.get() && data->hasKey("_id_")) {
      if ((*it)->getDriverName() == data->getStringValue("_id_")) {
        ret = (*it)->getDrvProperties();
        if (ret.get()) {
          ACULDBG_ << "get driver " << (*it)->getDriverName() << " property:" << ret->getJSONString();

        } else {
          ACULDBG_ << "get driver " << (*it)->getDriverName() << " ha NO property published";
        }
      }
    } else {
      ret = (*it)->getDrvProperties();
      if (ret.get()) {
        ACULDBG_ << "get driver " << (*it)->getDriverName() << " property:" << ret->getJSONString();

      } else {
        ACULDBG_ << "get driver " << (*it)->getDriverName() << " ha NO property published";
      }

      return ret;
    }
  }

  return ret;
}

void AbstractControlUnit::unitDefineDriver(std::vector<DrvRequestInfo>& neededDriver) {
  for (ControlUnitDriverListIterator iter = control_unit_drivers.begin();
       iter != control_unit_drivers.end();
       iter++) {
    //copy driver info to the system array reference
    neededDriver.push_back(*iter);
  }
}

void AbstractControlUnit::unitDefineCustomAttribute() {
  DrvMsg cmd;

  cmd.opcode                            = OpcodeType::OP_GET_PROPERTIES;
  int                               cnt = 0;
  chaos::common::data::CDataWrapper drv;

  for (VInstantitedDriverIterator it  = accessor_instances.begin(),
                                  end = accessor_instances.end();
       it != end;
       it++) {
    chaos::common::data::CDWUniquePtr res = (*it)->getDrvProperties();
    drv.addCSDataValue((*it)->getDriverName(), (res.get() ? *(res.get()) : chaos::common::data::CDataWrapper()));

    cnt++;
  }

 /* if (cnt) {
    //        drv.finalizeArrayForKey(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DRIVER_INFO);
    ACULDBG_ << " Adding driver properties to custom dataset " << chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DRIVER_INFO << ":" << drv.getJSONString();

    getAttributeCache()->addCustomAttribute(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DRIVER_INFO, drv);
    getAttributeCache()->setCustomAttributeValue(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DRIVER_INFO, drv);
  } else if (drv_info.get()) {
    ACULDBG_ << " Adding driver info to custom dataset " << chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_CU_INFO << ":" << drv_info->getJSONString();
    getAttributeCache()->addCustomAttribute(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_CU_INFO, *drv_info.get());
    getAttributeCache()->setCustomAttributeValue(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_CU_INFO, *drv_info.get());
  }*/
}

void AbstractControlUnit::_undefineActionAndDataset() {
  ACULDBG_ << "Remove Action Description";
  unitUndefineActionAndDataset();
}

void AbstractControlUnit::unitUndefineActionAndDataset() {
}

//! Get all managem declare action instance
void AbstractControlUnit::_getDeclareActionInstance(std::vector<const chaos::DeclareAction*>& declareActionInstance) {
  declareActionInstance.push_back(this);
}

//----------------------------------------- checklist method ------------------------------------------------
#pragma mark checklist method
void         AbstractControlUnit::doInitRpCheckList() {
  std::vector<std::string> attribute_names;
  push_errors   = 0;
  packet_lost   = 0;
  push_tot_size = 0;

  //rpc initialize service
  CHAOS_CHECK_LIST_START_SCAN_TO_DO(check_list_sub_service, "_init") {
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_INIT_STATE) {
      //call init sequence
      init(NULL);
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_INIT_SHARED_CACHE) {
      ACULDBG_ << "Allocate the user cache wrapper for:" + DatasetDB::getDeviceID();
      attribute_shared_cache_wrapper = new AttributeSharedCacheWrapper(attribute_value_shared_cache);

      ACULDBG_ << "Populating shared attribute cache for input attribute";
      DatasetDB::getDatasetAttributesName(DataType::Input, attribute_names);
      DatasetDB::getDatasetAttributesName(DataType::Bidirectional, attribute_names);
      initAttributeOnSharedAttributeCache(DOMAIN_INPUT, attribute_names);

      ACULDBG_ << "Populating shared attribute cache for output attribute";
      attribute_names.clear();
      DatasetDB::getDatasetAttributesName(DataType::Output, attribute_names);
      DatasetDB::getDatasetAttributesName(DataType::Bidirectional, attribute_names);
      initAttributeOnSharedAttributeCache(DOMAIN_OUTPUT, attribute_names);
      break;
    }

    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_COMPLETE_OUTPUT_ATTRIBUTE) {
      ACULDBG_ << "Complete shared attribute cache for output attribute";
      completeOutputAttribute();
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_COMPLETE_INPUT_ATTRIBUTE) {
      ACULDBG_ << "Complete shared attribute cache for input attribute";
      completeInputAttribute();
      break;
    }

    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_INIT_SYSTEM_CACHE) {
      ACULDBG_ << "Populating shared attribute cache for system attribute";
      initSystemAttributeOnSharedAttributeCache();
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_UNIT_DEFINE_ATTRIBUTE) {
    /*  std::string cu_load_param = getCUParam();

     if (isCUParamInJson()) {
        getAttributeCache()->addCustomAttribute(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_LOAD_PARAM, cu_load_param.size() + 1, chaos::DataType::TYPE_JSON);
      } else {
        getAttributeCache()->addCustomAttribute(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_LOAD_PARAM, cu_load_param.size() + 1, chaos::DataType::TYPE_STRING);
      }
      getAttributeCache()->setCustomAttributeValue(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_LOAD_PARAM, (void*)cu_load_param.c_str(), cu_load_param.size() + 1);

      //define the implementations custom variable
      AbstractControlUnit::unitDefineCustomAttribute();
      */
      unitDefineCustomAttribute();
      getAttributeCache()->setCustomDomainAsChanged();
      fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                                  cache_custom_attribute_vector);
      pushCustomDataset();
      
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_CREATE_FAST_ACCESS_CASCHE_VECTOR) {
      //create fast vector access for cached value
      fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT),
                            cache_output_attribute_vector);

      fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT),
                            cache_input_attribute_vector);

      fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM),
                            cache_system_attribute_vector);

      fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                            cache_custom_attribute_vector);
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_UNIT_INIT) {
      //initialize implementations
      unitInit();
      setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_lost", chaos::common::alarm::MultiSeverityAlarmLevelClear);
      setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_send_error", chaos::common::alarm::MultiSeverityAlarmLevelClear);

      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_UPDATE_CONFIGURATION) {
      //call update param function
      updateConfiguration(MOVE(init_configuration->clone()));
      //check if we need to do a restore on first start
      PropertyGroupShrdPtr cug = PropertyCollector::getGroup(chaos::ControlUnitPropertyKey::P_GROUP_NAME);
      if (cug.get()) {
        if (cug->hasProperty(chaos::ControlUnitPropertyKey::INIT_RESTORE_APPLY) &&
            cug->getProperty(chaos::ControlUnitPropertyKey::INIT_RESTORE_APPLY).getPropertyValue().isValid() &&
            cug->getProperty(chaos::ControlUnitPropertyKey::INIT_RESTORE_APPLY).getPropertyValue().asBool()) {
          //we need to add to stsart phase the restore one
          check_list_sub_service.getSharedCheckList("_start")->addElement(START_RPC_PHASE_RESTORE_ON_FIRST_START);
        }
      }

      if (init_configuration->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_PROP) && init_configuration->isVectorValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_PROP)) {
        chaos::common::data::CMultiTypeDataArrayWrapperSPtr ptr   = init_configuration->getVectorValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_PROP);
        std::map<std::string, std::string>                  props = ptr->toKVmap("name", "value");
        for (std::map<std::string, std::string>::iterator i = props.begin(); i != props.end(); i++) {
          ACULDBG_ << "Set CU property \"" << i->first << "\" to:" << i->second;

          chaos::common::data::Property<AbstractControlUnit>::setProperty(i->first, i->second,true);
        }
      }
      // try to call handler as setdataset attribute
      if (init_configuration->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DESCRIPTION)) {
        ChaosUniquePtr<chaos::common::data::CDataWrapper> dataset_object(init_configuration->getCSDataValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DESCRIPTION));
        if (dataset_object->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DESCRIPTION)) {
          CDWUniquePtr cdw_unique_ptr(new CDataWrapper());
          cu_ds_init.reset(new CDataWrapper());

          cdw_unique_ptr->addStringValue(NodeDefinitionKey::NODE_UNIQUE_ID, getCUID());
          //get the entity for device
          CMultiTypeDataArrayWrapperSPtr elementsDescriptions = dataset_object->getVectorValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DESCRIPTION);

          for (int idx = 0; idx < elementsDescriptions->size(); idx++) {
            //next element in dataset
            CDWUniquePtr elementDescription = elementsDescriptions->getCDataWrapperElementAtIndex(idx);
            //attribute name

            if (!elementDescription->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_ATTRIBUTE_NAME) ||
                !elementDescription->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_ATTRIBUTE_TYPE) ||
                !elementDescription->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DEFAULT_VALUE)) {
              continue;
            }
            string  attrName  = elementDescription->getStringValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_ATTRIBUTE_NAME);
            int32_t attrType  = elementDescription->getInt32Value(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_ATTRIBUTE_TYPE);
            string  attrValue = elementDescription->getStringValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_DEFAULT_VALUE);
            cu_ds_init->addCSDataValue(attrName, *elementDescription);
            if ((attrValue == "" || attrValue == "NA") && (attrType != DataType::TYPE_STRING)) {
              // not valid value not initialize
              ACULDBG_ << "skipped initialization of:" << attrName << " to:'" << attrValue << "' not valid for not string type";

              continue;
            }
            switch (attrType) {
              case DataType::TYPE_BOOLEAN:
                cdw_unique_ptr->addBoolValue(attrName, CDataVariant(attrValue).asBool());
                break;
              case DataType::TYPE_INT32:
                cdw_unique_ptr->addInt32Value(attrName, (int32_t)strtoul(attrValue.c_str(), 0, 0));
                break;
              case DataType::TYPE_INT64:
                cdw_unique_ptr->addInt64Value(attrName, (int64_t)strtoll(attrValue.c_str(), 0, 0));
                break;
              case DataType::TYPE_UINT64:
                cdw_unique_ptr->addUInt64Value(attrName, (uint64_t)strtoll(attrValue.c_str(), 0, 0));
                break;
              case DataType::TYPE_DOUBLE:
              case DataType::TYPE_FLOAT:
                cdw_unique_ptr->addDoubleValue(attrName, CDataVariant(attrValue).asDouble());
                break;
              case DataType::TYPE_STRING:
                cdw_unique_ptr->addStringValue(attrName, attrValue);
                break;
              case DataType::TYPE_JSON:
                cdw_unique_ptr->addStringValue(attrName, attrValue);
                break;
              case DataType::TYPE_BYTEARRAY:
                ACULERR_ << "Cannot Init Bytearray attribute :" << attrName;

                /*   CDataVariant v(attrValue);
                                    cdw_unique_ptr->addBinaryValue(attrName,
                                                                   v.asCDataBuffer()->getBuffer(),
                                                                   (uint32_t)v.asCDataBuffer()->getBufferSize());
                                                                   */
                break;
            }
          }

         /* if (cu_ds_init.get()) {
            ACULDBG_ << "INIT ATTRIBUTES:" << cu_ds_init->getJSONString();

            getAttributeCache()->addCustomAttribute(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_INITIALIZATION, *(cu_ds_init.get()));
            getAttributeCache()->setCustomAttributeValue(chaos::ControlUnitNodeDefinitionKey::CONTROL_UNIT_DATASET_INITIALIZATION, *(cu_ds_init.get()));

            fillCachedValueVector(attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM),
                                  cache_custom_attribute_vector);
          }*/
          CDWUniquePtr res = setDatasetAttribute(MOVE(cdw_unique_ptr));
        }
      }
      ChaosStringVector inp;
      ioTocheck.clear();
      limitTocheck.clear();
      attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT).getAttributeNames(inp);
      for (ChaosStringVector::iterator i = inp.begin(); i != inp.end(); i++) {
        setReadoutCheck(*i,true);        
      }
      inp.clear();
      attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT).getAttributeNames(inp);
      for (ChaosStringVector::iterator i = inp.begin(); i != inp.end(); i++) {
        setLimCheck(*i,true);
      }

      // set masks if any
      //* initialize alarm mask if any */
      if (init_configuration->hasKey(ControlUnitNodeDefinitionKey::CONTROL_UNIT_ALRM_MSK) && init_configuration->isVectorValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_ALRM_MSK)) {
        chaos::common::data::CMultiTypeDataArrayWrapperSPtr ptr  = init_configuration->getVectorValue(ControlUnitNodeDefinitionKey::CONTROL_UNIT_ALRM_MSK);
        std::map<std::string, std::string>                  msks = ptr->toKVmap("alarm", "mask");
        for (std::map<std::string, std::string>::iterator i = msks.begin(); i != msks.end(); i++) {
          setAlarmMask(i->first, atoi(i->second.c_str()));
        }
      }
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_init", INIT_RPC_PHASE_PUSH_DATASET) {
      //init on shared cache the all the dataaset with the default value
      //set first timestamp for simulate the run step
      int err;
      *timestamp_acq_cached_value->getValuePtr<uint64_t>() = TimingUtil::getTimeStamp();
      attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
      /*
                 *if ((err = pushSystemDataset()) != 0) {
                 ACULWRN_ << "cannot initialize system dataset, retrying";
                 sleep(1);
                 if ((err = pushSystemDataset()) != 0) {
                 throw CException(err, "cannot initialize system dataset", __PRETTY_FUNCTION__);
                 }
                 }
                 */
      if ((err = pushSystemDataset()) != 0) {
        throw CException(err, "cannot initialize system dataset (check live services)", __PRETTY_FUNCTION__);
      }
      /*if ((err = pushCUAlarmDataset()) != 0) {
                 ACULWRN_ << "cannot initialize CUAlarm dataset, retrying";
                 sleep(1);
                 if ((err = pushCUAlarmDataset()) != 0) {
                 throw CException(err, "cannot initialize CU alarm dataset", __PRETTY_FUNCTION__);
                 }
                 }*/
      if ((err = pushCUAlarmDataset()) != 0) {
        throw CException(err, "cannot initialize CU alarm dataset (check live services)", __PRETTY_FUNCTION__);
      }
      /*if ((err = pushDevAlarmDataset()) != 0) {
                 ACULWRN_ << "cannot initialize DEVAlarm dataset, retrying";
                 if ((err = pushDevAlarmDataset()) != 0) {
                 throw CException(err, "cannot initialize DEV alarm dataset", __PRETTY_FUNCTION__);
                 }
                 }*/
      if ((err = pushDevAlarmDataset()) != 0) {
        throw CException(err, "cannot initialize DEV alarm dataset (check live services)", __PRETTY_FUNCTION__);
      }
      attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT).markAllAsChanged();
      // if the CU can't push initial dataset is a real problem, we must detect immediately
      /* if ((err = pushOutputDataset()) != 0) {
                 ACULWRN_ << "cannot initialize output dataset, err:"<<err<<" retrying..";
                 sleep(1);
                 if ((err = pushOutputDataset()) != 0) {
                 throw CException(err, "cannot initialize output dataset", __PRETTY_FUNCTION__);
                 }
                 }*/

      if ((err = pushOutputDataset()) != 0) {
        throw CException(err, "cannot initialize output dataset (check live services)", __PRETTY_FUNCTION__);
      }
      attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT).markAllAsChanged();

      /*
                 if ((err = pushInputDataset()) != 0) {
                 ACULWRN_ << "cannot initialize input dataset, retrying";
                 sleep(1);
                 if ((err = pushInputDataset()) != 0) {
                 throw CException(err, "cannot initialize input dataset", __PRETTY_FUNCTION__);
                 }
                 }*/
      if ((err = pushInputDataset()) != 0) {
        throw CException(err, "cannot initialize input dataset (check live services)", __PRETTY_FUNCTION__);
      }

      attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM).markAllAsChanged();

      if ((err = pushCustomDataset()) != 0) {
        throw CException(err, "cannot initialize custom dataset (check live services)", __PRETTY_FUNCTION__);
      }
      break;
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_TO_DO(check_list_sub_service, "_init")
  
  //std::string command_queue = getDeviceID() + DataPackPrefixID::COMMAND_DATASET_POSTFIX;
  //DataManager::getInstance()->getDataLiveDriverNewInstance()->subscribe(command_queue);
  //DataManager::getInstance()->getDataLiveDriverNewInstance()->addHandler(command_queue, boost::bind(&AbstractControlUnit::consumerHandler, this, _1));
}
int AbstractControlUnit::subscribe(const std::string& key,bool subscribeon){
  int ret=chaos::common::message::MessagePSDriver::getConsumerDriver()->subscribe(key,subscribeon);
  chaos::common::message::MessagePSDriver::getConsumerDriver()->addHandler(key, boost::bind(&AbstractControlUnit::consumerHandler, this, _1),subscribeon);
  
  return ret;
}

int AbstractControlUnit::setReadoutCheck(const std::string& ioname,bool enable_disable){
        RangeValueInfo attributeInfo;
        if(enable_disable){
          if(ioTocheck.count(ioname)){
            ACULDBG_<<"check on "<<ioname<<" Already enabled";
            return 0;
          }
        } else {
          checkmap_t::iterator i=ioTocheck.find(ioname);
          if(i==ioTocheck.end()){
              ACULDBG_<<"check on "<<ioname<<" Already disabled";
              return 0;
          }
          ACULDBG_<<"check on "<<ioname<<" disabled";

          ioTocheck.erase(i);
          return ioTocheck.size();
        }
        
        getAttributeRangeValueInfo(ioname, attributeInfo);
        //ACULDBG_ << *i << " DOMAIN INPUT DIR:"<<attributeInfo.dir<<" warning:" << attributeInfo.warningThreshold<<" error:"<<attributeInfo.errorThreshold;

        if (attributeInfo.dir == chaos::DataType::Bidirectional) {
          //if(attributeInfo.)
          checkAttribute_t a;

          if (attributeInfo.warningThreshold.size() || attributeInfo.errorThreshold.size()) {
            std::stringstream ss;
            bool              error_set = false;
            ss << "Notify when '" << ioname << "'";
            if (attributeInfo.warningThreshold.size() > 0) {
              if (attributeInfo.warningThreshold == "0") {
                ss << " warning if set!=readout ,";

              } else {
                ss << " warning if |set-readout| > " << atof(attributeInfo.warningThreshold.c_str()) << ",";
              }
              error_set = true;
            }
            if (attributeInfo.errorThreshold.size() > 0) {
              if (attributeInfo.errorThreshold == "0") {
                ss << " error if set!=readout ,";

              } else {
                ss << " error if |set-readout| > " << atof(attributeInfo.errorThreshold.c_str());
              }
              error_set = true;
            }
            if (error_set) {
              addStateVariable(StateVariableTypeAlarmCU, ioname + "_out_of_set", ss.str());
            }

            a.i_idx = attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT).getIndexForName(ioname);
            a.o_idx = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT).getIndexForName(ioname)              ;
            a.range = attributeInfo;

            ioTocheck[ioname]=a;
            ACULDBG_ << "ENABLING CHECK " << ioTocheck.size() << " " << ioname << " set (" << a.i_idx << ") read (" << a.o_idx << ") :" << ss.str();
          }
        }
        return ioTocheck.size();
}
int AbstractControlUnit::setLimCheck(const std::string& ioname,bool enable_disable){
        RangeValueInfo attributeInfo;
        if(enable_disable){
          if(limitTocheck.count(ioname)){
            ACULDBG_<<"check limit on "<<ioname<<" Already enabled";
            return 0;
          }
        } else {
          checkmap_t::iterator i=limitTocheck.find(ioname);
          if(i==limitTocheck.end()){
              ACULDBG_<<"limit check on "<<ioname<<" Already disabled";
              return 0;
          }
          ACULDBG_<<"limit check on "<<ioname<<" disabled";

          limitTocheck.erase(i);
          return limitTocheck.size();
        }
        getAttributeRangeValueInfo(ioname, attributeInfo);

        bool limitcheck = false;
        if (attributeInfo.maxRange.size() || attributeInfo.warningMaxRange.size()) {
          std::stringstream ss;
          ss << " Error if " << ioname << " > " << attributeInfo.maxRange << ", warning if > " << attributeInfo.warningMaxRange;
          addStateVariable(StateVariableTypeAlarmCU, ioname + "_out_max_range", ss.str());
          limitcheck = true;
          ACULDBG_ << "ENABLING OUTPUT LIMIT CHECK " << limitTocheck.size() << " " << ss.str();
        }
        if (attributeInfo.minRange.size() || attributeInfo.warningMinRange.size()) {
          std::stringstream ss;
          ss << " Error if  " << ioname << " < " << attributeInfo.minRange << ", warning if > " << attributeInfo.warningMinRange;
          addStateVariable(StateVariableTypeAlarmCU, ioname + "_out_min_range", ss.str());
          limitcheck = true;
          ACULDBG_ << "ENABLING OUTPUT LIMIT CHECK " << limitTocheck.size() << " " << ss.str();
        }
        if (limitcheck) {
          checkAttribute_t a;
          a.range = attributeInfo;

          a.o_idx = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT).getIndexForName(ioname);

          a.i_idx = -1;
          limitTocheck[ioname]=a;
        }
        return limitTocheck.size();
}

std::string AbstractControlUnit::dsDefaultValue(const std::string& inputds) {
  chaos::common::data::RangeValueInfo attr_info;
  DatasetDB::getAttributeRangeValueInfo(inputds, attr_info);
  return attr_info.defaultValue;
}

void AbstractControlUnit::dsInitSetFromReadout() {
  for (int cnt = 0; cnt < cache_input_attribute_vector.size(); cnt++) {
    chaos::common::data::RangeValueInfo attr_info;
    std::string                         name = cache_input_attribute_vector[cnt]->name;
    DatasetDB::getAttributeRangeValueInfo(name, attr_info);
    if (attr_info.dir == chaos::DataType::Bidirectional) {
      if (attr_info.defaultValue.size() == 0) {
        for (int cntt = 0; cntt < cache_output_attribute_vector.size(); cntt++) {
          if (cache_output_attribute_vector[cntt]->name == name) {
            switch (attr_info.valueType) {
              case DataType::TYPE_BOOLEAN: {
                bool val = *cache_output_attribute_vector[cntt]->getValuePtr<bool>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_BOOLEAN to:" << val;

                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              case DataType::TYPE_DOUBLE: {
                double val = *cache_output_attribute_vector[cntt]->getValuePtr<double>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_DOUBLE to:" << val;
                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              case DataType::TYPE_FLOAT: {
                float val = *cache_output_attribute_vector[cntt]->getValuePtr<float>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_FLOAT to:" << val;
                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              case DataType::TYPE_INT32: {
                int32_t val = *cache_output_attribute_vector[cntt]->getValuePtr<int32_t>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_INT32 to:" << val;
                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              case DataType::TYPE_INT64: {
                int64_t val = *cache_output_attribute_vector[cntt]->getValuePtr<int64_t>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_INT64 to:" << val;
                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              case DataType::TYPE_STRING:{
                char* val = cache_output_attribute_vector[cntt]->getValuePtr<char>();
                ACULDBG_<< "Input '"<< name << "' Init TYPE_STRING to:" << val;
                cache_input_attribute_vector[cnt]->setValue(cache_output_attribute_vector[cntt]->value_buffer,cache_output_attribute_vector[cntt]->size);

                break;
              }
              default:
                ACULERR_ << name << " skipping initialization of complex type " << attr_info.valueType;

                break;
            }
          }
        }
      }
    }
  }
}

void AbstractControlUnit::doInitSMCheckList() {
  //rpc initialize service
  CHAOS_CHECK_LIST_START_SCAN_TO_DO(check_list_sub_service, "init") {
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "init", INIT_SM_PHASE_INIT_DB) {
      //cast to the CDatawrapper instance
      ACULDBG_ << "Initialize CU Database for device:" << DatasetDB::getDeviceID();
      run_id = CDW_GET_INT64_WITH_DEFAULT(init_configuration, ControlUnitNodeDefinitionKey::CONTROL_UNIT_RUN_ID, 0);
      DatasetDB::addAttributeToDataSetFromDataWrapper(*init_configuration);
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "init", INIT_SM_PHASE_CREATE_DATA_STORAGE) {
      //call init sequence
      //call update param function
      //initialize key data storage for device id
      //ACULDBG_ << "Create KeyDataStorage device:" << DatasetDB::getDeviceID();
      key_data_storage.reset(DataManager::getInstance()->getKeyDataStorageNewInstanceForKey(DatasetDB::getDeviceID()));
      aligned_ex_done=false;
      ACULDBG_ << "Call KeyDataStorage init implementation for deviceID:" << DatasetDB::getDeviceID() << " init:" << ((init_configuration.get()) ? init_configuration->getJSONString() : "NULL CONFIG");
      key_data_storage->init(init_configuration.get());
      break;
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_TO_DO(check_list_sub_service, "init")
}

void AbstractControlUnit::doStartRpCheckList() {
  CHAOS_CHECK_LIST_START_SCAN_TO_DO(check_list_sub_service, "_start") {
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_start", START_RPC_PHASE_IMPLEMENTATION) {
      start();
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_start", START_RPC_PHASE_UNIT) {
      unitStart();
      break;
    }
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "_start", START_RPC_PHASE_RESTORE_ON_FIRST_START) {
      try {
        checkForRestoreOnInit();
        check_list_sub_service.getSharedCheckList("_start")->removeElement(START_RPC_PHASE_RESTORE_ON_FIRST_START);
      } catch (CException& ex) {
        check_list_sub_service.getSharedCheckList("_start")->removeElement(START_RPC_PHASE_RESTORE_ON_FIRST_START);
        throw;
      }
      break;
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_TO_DO(check_list_sub_service, "_start")
}
void AbstractControlUnit::setDriverInfo(const chaos::common::data::CDataWrapper& info) {
  drv_info.reset(new CDataWrapper);
  info.copyAllTo(*drv_info.get());
  ACULDBG_ << "Set Driving info:" << drv_info->getJSONString();
}

void AbstractControlUnit::doStartSMCheckList() {
  CHAOS_CHECK_LIST_START_SCAN_TO_DO(check_list_sub_service, "start") {
    CHAOS_CHECK_LIST_DONE(check_list_sub_service, "start", START_SM_PHASE_STAT_TIMER) {
      //register timer for push statistic
      chaos::common::async_central::AsyncCentralManager::getInstance()->addTimer(this, 0, chaos::common::constants::CUTimersTimeoutinMSec);
      //get timestamp for first pushes metric acquisition
      last_push_rate_grap_ts = TimingUtil::getTimeStamp();
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_TO_DO(check_list_sub_service, "start")
}

void AbstractControlUnit::redoInitRpCheckList(bool throw_exception) {
  //rpc initialize service
  CHAOS_CHECK_LIST_START_SCAN_DONE(check_list_sub_service, "_init") {
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_INIT_STATE) {
      //saftely deinititalize the abstract control unit
      CHEK_IF_NEED_TO_THROW(throw_exception, deinit();)
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_INIT_SHARED_CACHE) {
      ACULDBG_ << "Deallocate the user cache wrapper for:" << DatasetDB::getDeviceID();
      if (attribute_shared_cache_wrapper) {
        delete (attribute_shared_cache_wrapper);
        attribute_shared_cache_wrapper = NULL;
      }
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_COMPLETE_OUTPUT_ATTRIBUTE) {
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_COMPLETE_INPUT_ATTRIBUTE) {
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_INIT_SYSTEM_CACHE) {
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_UNIT_DEFINE_ATTRIBUTE) {
      CHEK_IF_NEED_TO_THROW(throw_exception, unitDefineCustomAttribute();)
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_CREATE_FAST_ACCESS_CASCHE_VECTOR) {
      //clear all cache sub_structure
      CHEK_IF_NEED_TO_THROW(throw_exception,
                            cache_output_attribute_vector.clear();
                            cache_input_attribute_vector.clear();
                            cache_custom_attribute_vector.clear();
                            cache_system_attribute_vector.clear();)
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_CALL_UNIT_INIT) {
      ACULDBG_ << "Deinit custom deinitialization for device:" << DatasetDB::getDeviceID();
      CHEK_IF_NEED_TO_THROW(throw_exception, unitDeinit();)
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_UPDATE_CONFIGURATION) {
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_init", INIT_RPC_PHASE_PUSH_DATASET) {
      break;
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_DONE(check_list_sub_service, "_init")
}

void AbstractControlUnit::redoInitSMCheckList(bool throw_exception) {
  CHAOS_CHECK_LIST_START_SCAN_DONE(check_list_sub_service, "init") {
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "init", INIT_SM_PHASE_INIT_DB) {
      break;
    }
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "init", INIT_SM_PHASE_CREATE_DATA_STORAGE) {
      //remove key data storage
      CHEK_IF_NEED_TO_THROW(
          throw_exception,
          if (key_data_storage.get()) {
            ACULDBG_ << "Delete data storage driver for device:" << DatasetDB::getDeviceID();
            key_data_storage->deinit();
            key_data_storage.reset();
          })
      break;
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_DONE(check_list_sub_service, "init")
}

void AbstractControlUnit::redoStartRpCheckList(bool throw_exception) {
  CHAOS_CHECK_LIST_START_SCAN_DONE(check_list_sub_service, "_start") {
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_start", START_RPC_PHASE_IMPLEMENTATION) {
      CHEK_IF_NEED_TO_THROW(throw_exception, stop();)
      break;
    }
    //unit sto need to go after the abstract cu has been stopped
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "_start", START_RPC_PHASE_UNIT) {
      CHEK_IF_NEED_TO_THROW(throw_exception, unitStop();)
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_DONE(check_list_sub_service, "_start")
}

void AbstractControlUnit::redoStartSMCheckList(bool throw_exception) {
  CHAOS_CHECK_LIST_START_SCAN_DONE(check_list_sub_service, "start") {
    CHAOS_CHECK_LIST_REDO(check_list_sub_service, "start", START_SM_PHASE_STAT_TIMER) {
      //remove timer for push statistics
      CHEK_IF_NEED_TO_THROW(throw_exception, chaos::common::async_central::AsyncCentralManager::getInstance()->removeTimer(this);)
    }
  }
  CHAOS_CHECK_LIST_END_SCAN_DONE(check_list_sub_service, "start")
}

//----------------------------------------- protected initi/deinit method ------------------------------------------------
#pragma mark RPC State Machine method
CDWUniquePtr AbstractControlUnit::_init(CDWUniquePtr init_configuration) {
  if (getServiceState() == CUStateKey::INIT) {
    return CDWUniquePtr();
  }

  try {
    if (getServiceState() == CUStateKey::RECOVERABLE_ERROR) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -2, "Recoverable error state need to be recovered for %1%", % __PRETTY_FUNCTION__)
    }

    //if(getServiceState() != CUStateKey::DEINIT) throw CException(-1, DatasetDB::getDeviceID()+" need to be in deinit", __PRETTY_FUNCTION__);
    if (!attribute_value_shared_cache) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -3, "No Shared cache implementation found for %1%[%2%]", % DatasetDB::getDeviceID() % __PRETTY_FUNCTION__);
    }

    if (!SWEService::initImplementation(this, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Control Unit %1% can't be initilized [state mismatch]!", % DatasetDB::getDeviceID());
    }

  } catch (MetadataLoggingCException& ex) {
    throw;
  } catch (CException& ex) {
    throw MetadataLoggingCException(getCUID(),
                                    ex.errorCode,
                                    ex.errorMessage,
                                    ex.errorDomain);
  }

  try {
    //update configuraiton and own it
    this->init_configuration = MOVE(init_configuration);

    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_INITING,
                                                    true);
    doInitRpCheckList();

    //set healt to init
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_INIT,
                                                    true);
  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);
    //go in falta error
    SWEService::goInFatalError(this,
                               loggable_exception,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw loggable_exception;
  }
  return CDWUniquePtr();
}

CDWUniquePtr AbstractControlUnit::_start(CDWUniquePtr startParam) {
  hasstopped = false;

  if (getServiceState() == CUStateKey::START) {
    return CDWUniquePtr();
  }
  try {
    if (getServiceState() == CUStateKey::RECOVERABLE_ERROR) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Recoverable error state need to be recovered for %1%", % __PRETTY_FUNCTION__)
    }
    //call start method of the startable interface
    if (!SWEService::startImplementation(this, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Control Unit %1% can't be started [state mismatch]!", % DatasetDB::getDeviceID());
    }

  } catch (MetadataLoggingCException& ex) {
    throw;
  } catch (CException& ex) {
    throw MetadataLoggingCException(getCUID(),
                                    ex.errorCode,
                                    ex.errorMessage,
                                    ex.errorDomain);
  }

  try {
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_STARTING,
                                                    true);

    doStartRpCheckList();

    //set healt to start
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_START,
                                                    true);
  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);
    //go in falta error
    SWEService::goInFatalError(this, loggable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__);
    throw loggable_exception;
  }
  return CDWUniquePtr();
}

CDWUniquePtr AbstractControlUnit::_stop(CDWUniquePtr stopParam) {
  hasstopped = true;
  if (getServiceState() == CUStateKey::STOP) {
    ERR << " ALREADY IN STOP";
    return CDWUniquePtr();
  }
  try {
    /*
             * stop transition is possible with recoverable error
             *  if(getServiceState() == CUStateKey::RECOVERABLE_ERROR) {
             LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Recoverable error state need to be recovered for %1%", %__PRETTY_FUNCTION__)
             }
             */
    //first we start the deinitializaiton of the implementation unit
    if (!SWEService::stopImplementation(this, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Control Unit %1% can't be stopped [state mismatch]!", % DatasetDB::getDeviceID());
    }
  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    throw MetadataLoggingCException(getCUID(),
                                    ex.errorCode,
                                    ex.errorMessage,
                                    ex.errorDomain);
  }

  try {
    //set healt to start
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_STOPING,
                                                    true);
    /* {
  for (VInstantitedDriverIterator it  = accessor_instances.begin(),
                                  end = accessor_instances.end();
       it != end;
       it++) {
         // signal blocked queue
         if((*it)){
            (*it)->stop();
         }
  }
    }*/
    redoStartRpCheckList();
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_STOP,
                                                    true);
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_RATE,
                                                    0.0,
                                                    true);
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_SIZE,
                                                    0,
                                                    true);

  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);
    //go in falta error
    SWEService::goInFatalError(this, loggable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__);
    throw loggable_exception;
  }

  return CDWUniquePtr();
}

/*go
     deinit all datastorage
     */
CDWUniquePtr AbstractControlUnit::_deinit(CDWUniquePtr deinitParam) {
  if (getServiceState() == CUStateKey::DEINIT) {
    return CDWUniquePtr();
  }
  try {
    /*
             * deinit transition is possible with recoverable error
             *  if(getServiceState() == CUStateKey::RECOVERABLE_ERROR) {
             LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Recoverable error state need to be recovered for %1%", %__PRETTY_FUNCTION__)
             }
             */
    if (!SWEService::deinitImplementation(this, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
      LOG_AND_TROW_FORMATTED(ACULERR_, -1, "Control Unit %1% can't be deinitilized [state mismatch]!", % DatasetDB::getDeviceID());
    }
  } catch (MetadataLoggingCException& ex) {
    throw;
  } catch (CException& ex) {
    throw MetadataLoggingCException(getCUID(),
                                    ex.errorCode,
                                    ex.errorMessage,
                                    ex.errorDomain);
  }

  //first we start the deinitializaiton of the implementation unit
  try {
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_DEINITING,
                                                    true);

    redoInitRpCheckList();

    //set healt to deinit
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                    NodeHealtDefinitionValue::NODE_HEALT_STATUS_DEINIT,
                                                    true);
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_RATE,
                                                    0.0,
                                                    true);
    HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                    ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_SIZE,
                                                    0,
                                                    true);
  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);

    //go in falta error
    SWEService::goInFatalError(this, loggable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__);
    throw loggable_exception;
  }

  return CDWUniquePtr();
}

CDWUniquePtr AbstractControlUnit::_recover(CDWUniquePtr gdeinitParam) {
  if (getServiceState() != CUStateKey::RECOVERABLE_ERROR) throw MetadataLoggingCException(getCUID(), -1, DatasetDB::getDeviceID() + " need to be recoverable errore in the way to be recoverable!", __PRETTY_FUNCTION__);

  //first we start the deinitializaiton of the implementation unit
  try {
    if (SWEService::recoverError(this, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
    } else {
    }
  } catch (MetadataLoggingCException& ex) {
    SWEService::goInFatalError(this,
                               ex,
                               "AbstractControlUnit",
                               __PRETTY_FUNCTION__);
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);
    //go in falta error
    SWEService::goInFatalError(this, loggable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__);
    throw loggable_exception;
  }

  return CDWUniquePtr();
}

//! fill cache with found dataset at the restore point
void AbstractControlUnit::fillRestoreCacheWithDatasetFromTag(chaos::cu::data_manager::KeyDataStorageDomain domain,
                                                             CDataWrapper&                                 dataset,
                                                             AbstractSharedDomainCache&                    restore_cache) {
  std::vector<std::string> dataset_key;
  dataset.getAllKey(dataset_key);
  ChaosSharedPtr<SharedCacheLockDomain> w_lock = attribute_value_shared_cache->getLockOnDomain((SharedCacheDomain)domain, true);
  w_lock->lock();

  for (std::vector<std::string>::iterator it = dataset_key.begin();
       it != dataset_key.end();
       it++) {
    restore_cache.addAttribute((SharedCacheDomain)domain,
                               *it,
                               dataset.getVariantValue(*it));
  }
}

void AbstractControlUnit::checkForRestoreOnInit() {
  //now we can launch the restore the current input attrite, remeber that
  //input attribute are composed by mds so the type of restore data(static conf or live) is manage at mds leve
  //control unit in case off pply true need only to launch the restore on current input dataset set.
  try {
    ChaosUniquePtr<AttributeValueSharedCache> restore_cache(new AttributeValueSharedCache());
    if (!restore_cache.get()) throw MetadataLoggingCException(getCUID(), -3, "failed to allocate restore cache", __PRETTY_FUNCTION__);

    AttributeCache& ac_src = attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT);
    AttributeCache& ac_dst = restore_cache->getSharedDomain(DOMAIN_INPUT);
    ac_src.copyToAttributeCache(ac_dst);

    //unitRestoreToSnapshot
    if (unitRestoreToSnapshot(restore_cache.get())) {
      metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                      "Restore to initilization value run successfully");
    } else {
      metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelError,
                      "Restore to initilization value has fault");

      //input attribute already already updated
    }
  } catch (MetadataLoggingCException& ex) {
    throw;
  } catch (CException& ex) {
    MetadataLoggingCException loggable_exception(getCUID(),
                                                 ex.errorCode,
                                                 ex.errorMessage,
                                                 ex.errorDomain);

    DECODE_CHAOS_EXCEPTION(loggable_exception);
  }
}

/*!
     Restore the control unit to a precise tag
     */
CDWUniquePtr AbstractControlUnit::_unitRestoreToSnapshot(CDWUniquePtr restoreParam) {
  int err = 0;
  //check
  if (!restoreParam.get() || !restoreParam->hasKey(NodeDomainAndActionRPC::ACTION_NODE_RESTORE_PARAM_TAG)) return CDWUniquePtr();

  if (getServiceState() != CUStateKey::START) {
    throw MetadataLoggingCException(getCUID(), -1, "Control Unit restore can appen only in start state", __PRETTY_FUNCTION__);
  }

  if (!key_data_storage.get()) throw MetadataLoggingCException(getCUID(), -2, "Key data storage driver not allocated", __PRETTY_FUNCTION__);

  ChaosSharedPtr<AttributeValueSharedCache> restore_cache(new AttributeValueSharedCache());
  if (!restore_cache.get()) throw MetadataLoggingCException(getCUID(), -3, "failed to allocate restore cache", __PRETTY_FUNCTION__);

  ChaosSharedPtr<CDataWrapper> dataset_at_tag;
  //get tag alias
  const std::string restore_snapshot_tag = restoreParam->getStringValue(NodeDomainAndActionRPC::ACTION_NODE_RESTORE_PARAM_TAG);

  metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                  CHAOS_FORMAT("Start restoring snapshot tag for: %1%", % restore_snapshot_tag));

  attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_TAG)->setStringValue(restore_snapshot_tag, true, true);
  *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointRestoreStarted;  //start

  attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
  pushSystemDataset();

  //load snapshot to restore
  if ((err = key_data_storage->loadRestorePoint(restore_snapshot_tag))) {
    ACULERR_ << "Error loading dataset form snapshot tag: " << restore_snapshot_tag;
    *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointErrorOnInit;  //error

    attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
    pushSystemDataset();

    throw MetadataLoggingCException(getCUID(), err, "Error loading dataset form snapshot", __PRETTY_FUNCTION__);
  } else {
    restore_cache->init(NULL);

    for (int idx = 0; idx < 4; idx++) {
      //dataset loading sucessfull
      dataset_at_tag = key_data_storage->getDatasetFromRestorePoint(restore_snapshot_tag,
                                                                    (KeyDataStorageDomain)idx);
      if (dataset_at_tag.get()) {
        //fill cache with dataset key/value
        fillRestoreCacheWithDatasetFromTag((KeyDataStorageDomain)idx,
                                           *dataset_at_tag.get(),
                                           *restore_cache.get());
        ACULDBG_ << CHAOS_FORMAT("Dataset restored from tag %1% -> %2%: %3%", % restore_snapshot_tag % datasetTypeToHuman(idx) % dataset_at_tag->getJSONString());
      }
    }
    *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointRestoreRunning;  //running
    attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
    pushSystemDataset();

    try {
      //unitRestoreToSnapshot
      if (unitRestoreToSnapshot(restore_cache.get())) {
        metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                        CHAOS_FORMAT("Restore for %1% has been run successfully", % restore_snapshot_tag));
        *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointRestoreReached;  //end
        attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
        pushSystemDataset();

      } else {
        metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelError,
                        CHAOS_FORMAT("Restore for %1% has been faulted", % restore_snapshot_tag));

        *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointErrorOnStart;  //error
        attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
        pushSystemDataset();
        //input dataset need to be update
        AttributeCache& ac_src = restore_cache->getSharedDomain(DOMAIN_INPUT);
        AttributeCache& ac_dst = attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT);
        ac_src.copyToAttributeCache(ac_dst);
      }
    } catch (MetadataLoggingCException& ex) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointErrorOnRunning;  //error
      attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
      pushSystemDataset();
      throw;
    } catch (CException& ex) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::SETPOINT_STATE)->getValuePtr<int32_t>() = ControlUnitNodeDefinitionType::SetpointErrorException;  //error
      attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
      pushSystemDataset();
      MetadataLoggingCException loggable_exception(getCUID(),
                                                   ex.errorCode,
                                                   ex.errorMessage,
                                                   ex.errorDomain);

      DECODE_CHAOS_EXCEPTION(loggable_exception);
    }

    //end
    try {
      restore_cache->deinit();
    } catch (MetadataLoggingCException& ex) {
      throw;
    } catch (CException& ex) {
      MetadataLoggingCException loggable_exception(getCUID(),
                                                   ex.errorCode,
                                                   ex.errorMessage,
                                                   ex.errorDomain);
      DECODE_CHAOS_EXCEPTION(loggable_exception);
    } catch (...) {
      ACULERR_ << "General error on deinit restore cache";
    }

    //! clear snapshoted dataset to free memeory
    key_data_storage->clearRestorePoint(restore_snapshot_tag);
  }
  metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                  CHAOS_FORMAT("End restoring snapshot tag for: %1%", % restore_snapshot_tag));
  return CDWUniquePtr();
}

/*
     Receive the event for set the dataset input element
     */
CDWUniquePtr AbstractControlUnit::_setDatasetAttribute(CDWUniquePtr dataset_attribute_values) {
  CDWUniquePtr result;
  try {
    if (!dataset_attribute_values.get()) {
      throw MetadataLoggingCException(getCUID(), -1, "No Input parameter", __PRETTY_FUNCTION__);
    }
    if (SWEService::getServiceState() == CUStateKey::DEINIT) {
      throw MetadataLoggingCException(getCUID(), -3, "The Control Unit is in deinit state", __PRETTY_FUNCTION__);
    }
    ACULDBG_<<"Setdataset:"<<dataset_attribute_values->getJSONString();
    //send dataset attribute change pack to control unit implementation
    result = setDatasetAttribute(MOVE(dataset_attribute_values));
  } catch (CException& ex) {
    throw;
  }
  return result;
}
#pragma mark State Machine method
// Startable Service method
void AbstractControlUnit::init(void* init_data) {
  //allocate metadata loggin channel for alarm
  alarm_logging_channel = (AlarmLoggingChannel*)MetadataLoggingManager::getInstance()->getChannel("AlarmLoggingChannel");
  if (alarm_logging_channel == NULL) {
    LOG_AND_TROW(ACULERR_, -1, "Alarm logging channel not found");
  }

  standard_logging_channel = (StandardLoggingChannel*)MetadataLoggingManager::getInstance()->getChannel("StandardLoggingChannel");
  if (standard_logging_channel == NULL) {
    LOG_AND_TROW(ACULERR_, -2, "Standard logging channel not found");
  }
  standard_logging_channel->setLogLevel(common::metadata_logging::StandardLoggingChannel::LogLevelInfo);
  //the init of the implementation unit goes after the infrastructure one
  doInitSMCheckList();
}

// Startable Service method
void AbstractControlUnit::start() {
  setStateVariableSeverity(StateVariableTypeAlarmDEV,chaos::common::alarm::MultiSeverityAlarmLevelClear);
  setStateVariableSeverity(StateVariableTypeAlarmCU,chaos::common::alarm::MultiSeverityAlarmLevelClear);
  doStartSMCheckList();
}

// Startable Service method
void AbstractControlUnit::stop() {

  setStateVariableSeverity(StateVariableTypeAlarmDEV,chaos::common::alarm::MultiSeverityAlarmLevelClear);
  setStateVariableSeverity(StateVariableTypeAlarmCU,chaos::common::alarm::MultiSeverityAlarmLevelClear);
  redoStartSMCheckList();
}

// Startable Service method
void AbstractControlUnit::deinit() {
  redoInitSMCheckList();
  if (alarm_logging_channel) {
    MetadataLoggingManager::getInstance()->releaseChannel(alarm_logging_channel);
    alarm_logging_channel = NULL;
  }

  if (standard_logging_channel) {
    MetadataLoggingManager::getInstance()->releaseChannel(standard_logging_channel);
    standard_logging_channel = NULL;
  }
}

//! State machine is gone into recoverable error
void AbstractControlUnit::recoverableErrorFromState(int last_state, chaos::CException& ex) {
  ACULERR_ << "recoverableErrorFromState with state:" << last_state;

  //update healt tstatus to report recoverable error
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  NodeHealtDefinitionValue::NODE_HEALT_STATUS_RERROR,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_CODE,
                                                  ex.errorCode,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_MESSAGE,
                                                  ex.errorMessage,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_DOMAIN,
                                                  ex.errorDomain,
                                                  true);
  //we stop the run action
  CHAOS_NOT_THROW(stop();)
}

//! State machine is gone into recoverable error
bool AbstractControlUnit::beforeRecoverErrorFromState(int last_state) {
  ACULERR_ << "beforeRecoverErrorFromState with state:" << last_state;
  std::string last_state_str;
  switch (last_state) {
    case CUStateKey::INIT:
      last_state_str = NodeHealtDefinitionValue::NODE_HEALT_STATUS_INIT;
      break;
    case CUStateKey::DEINIT:
      last_state_str = NodeHealtDefinitionValue::NODE_HEALT_STATUS_DEINIT;
      break;
    case CUStateKey::START:
      last_state_str = NodeHealtDefinitionValue::NODE_HEALT_STATUS_START;
      break;
    case CUStateKey::STOP:
      last_state_str = NodeHealtDefinitionValue::NODE_HEALT_STATUS_STOP;
      break;
  }
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  last_state_str,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_CODE,
                                                  0,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_MESSAGE,
                                                  "",
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_DOMAIN,
                                                  "",
                                                  true);
  //restart the cotnrol unit
  CHAOS_NOT_THROW(start();)
  return true;
}

//! State machine is gone into recoverable error
void AbstractControlUnit::recoveredToState(int last_state) {
  ACULERR_ << "recoveredToState with state:" << last_state;
}

//! State machine is gone into an unrecoverable error
void AbstractControlUnit::fatalErrorFromState(int last_state, chaos::CException& ex) {
  ACULERR_ << "fatalErrorFromState with state:" << last_state;
  // signal before
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  NodeHealtDefinitionValue::NODE_HEALT_STATUS_FERROR,
                                                  true);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_CODE,
                                                  ex.errorCode,
                                                  true);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_MESSAGE,
                                                  ex.errorMessage,
                                                  true);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_DOMAIN,
                                                  ex.errorDomain,
                                                  true);
  switch (last_state) {
    case CUStateKey::INIT:
      //deinit
      //CHAOS_NOT_THROW(redoInitSMCheckList(false);)
      deinit();
      CHAOS_NOT_THROW(redoInitRpCheckList(false);)
      break;
    case CUStateKey::DEINIT:

      break;
    case CUStateKey::START:
      //stop
      //CHAOS_NOT_THROW(redoStartSMCheckList(false);)
      stop();
      CHAOS_NOT_THROW(redoStartRpCheckList(false);)
      //deinit
      //CHAOS_NOT_THROW(redoInitSMCheckList(false);)
      /*            deinit();
                CHAOS_NOT_THROW(redoInitRpCheckList(false);)*/
      break;
    case CUStateKey::STOP:
      break;
  }  //update healt tstatus to report recoverable error
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  NodeHealtDefinitionValue::NODE_HEALT_STATUS_FERROR,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_CODE,
                                                  ex.errorCode,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_MESSAGE,
                                                  ex.errorMessage,
                                                  false);
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_LAST_ERROR_DOMAIN,
                                                  ex.errorDomain,
                                                  true);
}

void AbstractControlUnit::fillCachedValueVector(AttributeCache&               attribute_cache,
                                                std::vector<AttributeValue*>& cached_value) {
  cached_value.clear();
  for (int idx = 0;
       idx < attribute_cache.getNumberOfAttributes();
       idx++) {
    cached_value.push_back(attribute_cache.getValueSettingForIndex(idx));
  }
}
int AbstractControlUnit::checkLimFn(double rval, const chaos::common::data::RangeValueInfo& i, int dir) {
  std::string alrmmax = i.name + ((dir == 0) ? "_in_max_range" : "_out_max_range");
  std::string alrmmin = i.name + ((dir == 0) ? "_in_min_range" : "_out_min_range");
  int         err     = 0;
  if (i.warningMaxRange.size()) {
    double max = atof(i.warningMaxRange.c_str());
    if (rval > max) {
      chaos::common::alarm::MultiSeverityAlarmLevel l;
      getStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, l);
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, chaos::common::alarm::MultiSeverityAlarmLevelWarning);
      if (l != chaos::common::alarm::MultiSeverityAlarmLevelWarning) {
        ACULDBG_ << i.name << " WARN MAX CHECK failed " << rval << " > " << max << " alrm:" << alrmmax;
      }
      err++;
    } else {
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, chaos::common::alarm::MultiSeverityAlarmLevelClear);
    }
  }
  if (i.maxRange.size()) {
    double max = atof(i.maxRange.c_str());
    if (rval > max) {
      chaos::common::alarm::MultiSeverityAlarmLevel l;
      getStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, l);
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, chaos::common::alarm::MultiSeverityAlarmLevelHigh);
      if (l != chaos::common::alarm::MultiSeverityAlarmLevelHigh) {
        ACULDBG_ << i.name << " MAX CHECK failed " << rval << " > " << max << " alrm:" << alrmmax;
      }
      err++;

    } else {
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmax, chaos::common::alarm::MultiSeverityAlarmLevelClear);
    }
  }
  if (i.warningMinRange.size()) {
    double min = atof(i.warningMinRange.c_str());
    if (rval < min) {
      chaos::common::alarm::MultiSeverityAlarmLevel l;
      getStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, l);
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, chaos::common::alarm::MultiSeverityAlarmLevelWarning);
      if (l != chaos::common::alarm::MultiSeverityAlarmLevelWarning) {
        ACULDBG_ << i.name << " WARN MIN CHECK failed " << rval << " < " << min << " alrm:" << alrmmin;
      }
      err++;

    } else {
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, chaos::common::alarm::MultiSeverityAlarmLevelClear);
    }
  }
  if (i.minRange.size()) {
    double min = atof(i.minRange.c_str());
    if (rval < min) {
      chaos::common::alarm::MultiSeverityAlarmLevel l;
      getStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, l);
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, chaos::common::alarm::MultiSeverityAlarmLevelHigh);
      if (l != chaos::common::alarm::MultiSeverityAlarmLevelHigh) {
        ACULDBG_ << i.name << " MIN CHECK failed " << rval << " < " << min << " alrm:" << alrmmin;
      }
      err++;

    } else {
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrmmin, chaos::common::alarm::MultiSeverityAlarmLevelClear);
    }
  }
  return err;
}
int AbstractControlUnit::checkFn(double sval, double rval, const chaos::common::data::RangeValueInfo& i) {
  int         err  = 0;
  std::string alrm = i.name + "_out_of_set";

  //  checkLimFn(rval,i);

  double res = fabs(sval - rval);
 // ACULAPP_ << i.name << " CHECK :" << rval << " setpoint :" << sval << " threshold:" << atof(i.errorThreshold.c_str()) << " res:" << res;

  if (i.errorThreshold.size()) {
    if (i.errorThreshold.c_str() == "0") {
      if (sval != rval) {
        ACULDBG_ << i.name << " |SETPOINT-READOUT| ERROR CHECK failed " << rval << " != " << sval;

        setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, chaos::common::alarm::MultiSeverityAlarmLevelHigh);
        err++;
        return err;
      }
    } else if (res > atof(i.errorThreshold.c_str())) {
      ACULDBG_ << i.name << " |SETPOINT-READOUT| ERROR CHECK failed readout:" << rval << " setpoint :" << sval << " threshold:" << atof(i.errorThreshold.c_str()) << " res:" << res;

      setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, chaos::common::alarm::MultiSeverityAlarmLevelHigh);
      err++;
      return err;
    }
  }
  if (i.warningThreshold.size()) {
    if (i.warningThreshold.c_str() == "0") {
      if (sval != rval) {
        ACULDBG_ << i.name << " |SETPOINT-READOUT| WARNING CHECK failed " << rval << " != " << sval;

        setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, chaos::common::alarm::MultiSeverityAlarmLevelWarning);
        err++;
        return err;
      }
    } else if (res > atof(i.warningThreshold.c_str())) {
      ACULDBG_ << i.name << " |SETPOINT-READOUT| WARNING CHECK failed readout:" << rval << " setpoint:" << sval << " threshold:" << atof(i.warningThreshold.c_str()) << " res:" << res;
      setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, chaos::common::alarm::MultiSeverityAlarmLevelWarning);
      err++;
      return err;
    }
  }
  //ACULAPP_ << i.name << " CLEAR CHECK :" << rval << " setpoint :" << sval << " threshold:" << atof(i.errorThreshold.c_str()) << " res:" << res;

  setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, chaos::common::alarm::MultiSeverityAlarmLevelClear);

  return err;
}
int AbstractControlUnit::checkAlarms() {
  return checkStdAlarms();
}

int AbstractControlUnit::checkStdAlarms() {
  checkmap_t::iterator i;
  int                                     alarms = 0;
  //ACULDBG_ << "CHECK TO PERFORM :"<<ioTocheck.size();
  for (i = ioTocheck.begin(); i != ioTocheck.end(); i++) {
    int level = 0;
    //std::string alrm  = i->range.name + "_out_of_set";
    //   setStateVariableSeverity(StateVariableTypeAlarmCU, alrm, (common::alarm::MultiSeverityAlarmLevel)2);
    // ACULDBG_ << "NAME :"<<i->range.name<<" INPUT VAL:"<<i->input<<" OUTPUT VAL:"<<i->output;
    // ACULDBG_ << "INAME :"<<i->input->name<<" ONAME :"<<i->output->name;
    double          res;
    AttributeValue* inp = cache_input_attribute_vector[i->second.i_idx];
    AttributeValue* out = cache_output_attribute_vector[i->second.o_idx];
    /* {
      CDataWrapper t;
      fillCDatawrapperWithCachedValue(cache_input_attribute_vector,t);
    }*/
   //   ACULDBG_ << i->first << " |SETPOINT-READOUT|CHECKING..." << " threshold:" << atof(i->second.range.warningThreshold.c_str());

    if (inp == NULL || out == NULL) {
      continue;
    }
    switch (inp->type) {
      case DataType::TYPE_BOOLEAN: {
        int32_t sval    = (int32_t)*(inp->getValuePtr<bool>());
        int32_t rval = (int32_t)*(out->getValuePtr<bool>());
       //   ACULDBG_ << i->first << " |SETPOINT-READOUT|CHECK readout:" << rval << " setpoint:" << sval << " threshold:" << atof(i->second.range.warningThreshold.c_str());

        level        = checkFn(sval, rval, i->second.range);

        break;
      }
      case DataType::TYPE_INT32: {
        int32_t sval = *(inp->getValuePtr<int32_t>());
        int32_t rval = *(out->getValuePtr<int32_t>());
        level        = checkFn(sval, rval, i->second.range);

        break;
      }
      case DataType::TYPE_INT64: {
        int64_t sval = *(inp->getValuePtr<int64_t>());
        int64_t rval = *(out->getValuePtr<int64_t>());
        level        = checkFn(sval, rval, i->second.range);

        break;
      }
      case DataType::TYPE_DOUBLE: {
        double sval = *(inp->getValuePtr<double>());
        double rval = *(out->getValuePtr<double>());
        //      ACULDBG_ << "INPUT:"<<inp->name<<" val:"<<sval;

        level = checkFn(sval, rval, i->second.range);

        break;
      }
      case DataType::TYPE_FLOAT: {
        float sval = *(inp->getValuePtr<float>());
        float rval = *(out->getValuePtr<float>());
        //      ACULDBG_ << "INPUT:"<<inp->name<<" val:"<<sval;

        level = checkFn(sval, rval, i->second.range);

        break;
      }
      default:
        break;
    }

    alarms += level;
  }

  for (i = limitTocheck.begin(); i != limitTocheck.end(); i++) {
    int    level = 0;
    double res;
    
    AttributeValue* val = NULL;
    int             dir = -1;
    if (i->second.i_idx >= 0) {
      dir = 0;
      val = cache_input_attribute_vector[i->second.i_idx];
    } else if (i->second.o_idx >= 0) {
      dir = 1;
      val = cache_output_attribute_vector[i->second.o_idx];
    }
    if (val) {
      switch (val->type) {
        case DataType::TYPE_INT32: {
          int32_t sval = *(val->getValuePtr<int32_t>());
          level        = checkLimFn(sval, i->second.range, dir);

          break;
        }
        case DataType::TYPE_INT64: {
          int64_t sval = *(val->getValuePtr<int64_t>());
          level        = checkLimFn(sval, i->second.range, dir);

          break;
        }
        case DataType::TYPE_DOUBLE: {
          double sval = *(val->getValuePtr<double>());
          level       = checkLimFn(sval, i->second.range, dir);

          break;
        }
        case DataType::TYPE_FLOAT: {
          float sval = *(val->getValuePtr<float>());
          level       = checkLimFn(sval, i->second.range, dir);

          break;
        }
        default:
          break;
      }
    }

    alarms += level;
  }
  return alarms;
}

void AbstractControlUnit::initAttributeOnSharedAttributeCache(SharedCacheDomain         domain,
                                                              std::vector<std::string>& attribute_names) {
  //add input attribute to shared setting
  CHAOS_ASSERT(attribute_value_shared_cache)
  RangeValueInfo  attributeInfo;
  AttributeCache& attribute_setting = attribute_value_shared_cache->getSharedDomain(domain);

  for (int idx = 0;
       idx < attribute_names.size();
       idx++) {
    attributeInfo.reset();

    // retrive the attribute description from the device database
    DatasetDB::getAttributeRangeValueInfo(attribute_names[idx], attributeInfo);
    // add the attribute to the shared setting object
    attribute_setting.addAttribute(attribute_names[idx], attributeInfo.maxSize, attributeInfo.valueType, attributeInfo.binType);

    if (!attributeInfo.defaultValue.size()) continue;

    //setting value using index (the index into the sharedAttributeSetting are sequencial to the inserted order)
    try {
      switch (attributeInfo.valueType) {
        case DataType::TYPE_BOOLEAN: {
          bool val = boost::lexical_cast<bool>(attributeInfo.defaultValue);
          attribute_setting.setValueForAttribute(idx, &val, sizeof(bool));
          ACULDBG_ << domain << " Init TYPE_BOOLEAN attribute:'" << attribute_names[idx] << "' to:" << val;
          break;
        }
        case DataType::TYPE_DOUBLE: {
          double val = boost::lexical_cast<double>(attributeInfo.defaultValue);
          attribute_setting.setValueForAttribute(idx, &val, sizeof(double));
          ACULDBG_ << domain << " Init TYPE_DOUBLE attribute:'" << attribute_names[idx] << "' to:" << val;
          break;
        }
        case DataType::TYPE_FLOAT: {
          float val = boost::lexical_cast<float>(attributeInfo.defaultValue);
          attribute_setting.setValueForAttribute(idx, &val, sizeof(float));
          ACULDBG_ << domain << " Init TYPE_FLOAT attribute:'" << attribute_names[idx] << "' to:" << val;
          break;
        }
        case DataType::TYPE_INT32: {
          int32_t val = strtoul(attributeInfo.defaultValue.c_str(), 0, 0);  //boost::lexical_cast<int32_t>(attributeInfo.defaultValue);
          attribute_setting.setValueForAttribute(idx, &val, sizeof(int32_t));
          ACULDBG_ << domain << " Init TYPE_INT32 attribute:'" << attribute_names[idx] << "' to:" << val << " 0x" << std::hex << val << std::dec;
          break;
        }
        case DataType::TYPE_INT64: {
          int64_t val = strtoll(attributeInfo.defaultValue.c_str(), 0, 0);  //boost::lexical_cast<int64_t>(attributeInfo.defaultValue);
          attribute_setting.setValueForAttribute(idx, &val, sizeof(int64_t));
          ACULDBG_ << domain << " Init TYPE_INT64 attribute:'" << attribute_names[idx] << "' to:" << val << " 0x" << std::hex << val << std::dec;
          break;
        }
        case DataType::TYPE_JSON: {
          CDataWrapper tmp;
          tmp.setSerializedJsonData(attributeInfo.defaultValue.c_str());
          attribute_setting.setValueForAttribute(idx, tmp);
          ACULDBG_ << domain << " Init  TYPE_CLUSTER attribute:'" << attribute_names[idx] << "' to:" << attributeInfo.defaultValue.c_str();
          break;
        }
        case DataType::TYPE_STRING: {
          const char* val = attributeInfo.defaultValue.c_str();
          attribute_setting.setValueForAttribute(idx, val, (uint32_t)attributeInfo.defaultValue.size());
          ACULDBG_ << domain << " Init  TYPE_STRING attribute:" << attribute_names[idx] << " to:" << val;
          break;
        }
        case DataType::TYPE_BYTEARRAY: {
          ACULDBG_ << domain << " Binary default setting attribute:'" << attribute_names[idx] << "' size:" << (uint32_t)attributeInfo.defaultValue.size() << " has not been yet managed";
          break;
        }
        default:
          ACULERR_ << domain << " Uknown attribute type, attribute:'" << attribute_names[idx] << "' size:" << (uint32_t)attributeInfo.defaultValue.size() << " has not been yet managed";

          break;
      }
    } catch (boost::bad_lexical_cast const& e) {
      ACULERR_ << e.what();
    }
  }
}

void AbstractControlUnit::completeOutputAttribute() {
  ACULDBG_ << "Complete the shared cache output attribute";
  AttributeCache& domain_attribute_setting = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT);

  //add timestamp
  domain_attribute_setting.addAttribute(DataPackCommonKey::DPCK_TIMESTAMP, sizeof(uint64_t), DataType::TYPE_INT64);
  timestamp_acq_cached_value = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(DataPackCommonKey::DPCK_TIMESTAMP));
  if(use_custom_high_resolution_timestamp){
    domain_attribute_setting.addAttribute(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP, sizeof(uint64_t), DataType::TYPE_INT64);
    timestamp_hw_acq_cached_value = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP));
  }
}

void AbstractControlUnit::completeInputAttribute() {
}

AbstractSharedDomainCache* AbstractControlUnit::_getAttributeCache() {
  return attribute_value_shared_cache;
}
int AbstractControlUnit::incomingMessage(const std::string& key,  chaos::common::data::CDWUniquePtr& data) {
  ACULDBG_ << "message from :" << key << " data:" << (data.get() ? data->getJSONString() : "NONE");
  return 0;
}

void AbstractControlUnit::initSystemAttributeOnSharedAttributeCache() {
  AttributeCache& domain_attribute_setting = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);

  //add schedule time
  ACULDBG_ << "Adding system attribute on shared cache";
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::THREAD_SCHEDULE_DELAY, 0, DataType::TYPE_INT64);
  thread_schedule_daly_cached_value = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(ControlUnitDatapackSystemKey::THREAD_SCHEDULE_DELAY));

  //add busy state
  domain_attribute_setting.addAttribute("busy", 0, DataType::TYPE_BOOLEAN);
  busy=false;
  

  //add bypass state
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::BYPASS_STATE, 0, DataType::TYPE_BOOLEAN);

  //add AlarmCU state
  domain_attribute_setting.addAttribute(chaos::ControlUnitDatapackSystemKey::CU_ALRM_LEVEL, 0, DataType::TYPE_INT32);

  //add AlarmDev state
  domain_attribute_setting.addAttribute(chaos::ControlUnitDatapackSystemKey::DEV_ALRM_LEVEL, 0, DataType::TYPE_INT32);

  //add burst operation state
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::BURST_STATE, 0, DataType::TYPE_BOOLEAN);

  //add burst operation state
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::BURST_CNT_DOWN, 0, DataType::TYPE_INT32);

  //add burst operation tag
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::BURST_TAG, 0, DataType::TYPE_STRING);

  //add setpoint operation state
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::SETPOINT_STATE, 0, DataType::TYPE_INT32);

  //add setpoint operation tag
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::SETPOINT_TAG, 0, DataType::TYPE_STRING);

  //add storage type
  domain_attribute_setting.addAttribute(DataServiceNodeDefinitionKey::DS_STORAGE_TYPE, 0, DataType::TYPE_INT32);

  //add live time
  domain_attribute_setting.addAttribute(DataServiceNodeDefinitionKey::DS_STORAGE_LIVE_TIME, 0, DataType::TYPE_INT64);

  //add history time
  domain_attribute_setting.addAttribute(DataServiceNodeDefinitionKey::DS_STORAGE_HISTORY_TIME, 0, DataType::TYPE_INT64);
  domain_attribute_setting.addAttribute(DataServiceNodeDefinitionKey::DS_STORAGE_LOG_TIME, 0, DataType::TYPE_INT64);

  //add update anyway
  domain_attribute_setting.addAttribute(DataServiceNodeDefinitionKey::DS_UPDATE_ANYWAY, DS_UPDATE_ANYWAY_DEF, DataType::TYPE_INT32);

  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::CU_LOG_MAX_MS, CONTROL_UNIT_LOG_MAX_MS_DEF, DataType::TYPE_INT32);

  //command status
  domain_attribute_setting.addAttribute(DataPackSystemKey::DP_SYS_QUEUED_CMD, 0, DataType::TYPE_INT32);
  domain_attribute_setting.addAttribute(DataPackSystemKey::DP_SYS_STACK_CMD, 0, DataType::TYPE_INT32);
  domain_attribute_setting.addAttribute(ControlUnitDatapackSystemKey::RUNNING_COMMAND_ALIAS, 0, DataType::TYPE_STRING);
  // add hostname
  std::string hn = chaos::GlobalConfiguration::getInstance()->getHostname();
  domain_attribute_setting.addAttribute(chaos::ControlUnitDatapackSystemKey::CU_SOURCE_HOSTNAME, (uint32_t)hn.size(), DataType::TYPE_STRING);
  char* str_ptr = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(chaos::ControlUnitDatapackSystemKey::CU_SOURCE_HOSTNAME))->getValuePtr<char>();
  strncpy(str_ptr, hn.c_str(), hn.size());
  // add class name
  hn = getCUClass();

  domain_attribute_setting.addAttribute(chaos::ControlUnitDatapackSystemKey::CU_CLASS_TYPE, (uint32_t)hn.size(), DataType::TYPE_STRING);
  str_ptr = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(chaos::ControlUnitDatapackSystemKey::CU_CLASS_TYPE))->getValuePtr<char>();
  strncpy(str_ptr, hn.c_str(), hn.size());

  //add unit type
  domain_attribute_setting.addAttribute(DataPackSystemKey::DP_SYS_UNIT_TYPE, (uint32_t)control_unit_type.size(), DataType::TYPE_STRING);
  str_ptr = domain_attribute_setting.getValueSettingForIndex(domain_attribute_setting.getIndexForName(DataPackSystemKey::DP_SYS_UNIT_TYPE))->getValuePtr<char>();
  strncpy(str_ptr, control_unit_type.c_str(), control_unit_type.size());
}

/*
     Get the current control unit state
     */
CDWUniquePtr AbstractControlUnit::_getState(CDWUniquePtr getStatedParam) {
  CreateNewDataWrapper(result, );
  result->addInt32Value(CUStateKey::CONTROL_UNIT_STATE, static_cast<CUStateKey::ControlUnitState>(SWEService::getServiceState()));
  return result;
}
chaos::CUStateKey::ControlUnitState AbstractControlUnit::getState() {
  return static_cast<CUStateKey::ControlUnitState>(SWEService::getServiceState());
}

CDWUniquePtr AbstractControlUnit::_submitStorageBurst(CDWUniquePtr data) {
  common::data::structured::DatasetBurstSDWrapper db_sdw;
  db_sdw.deserialize(data.get());
  DatasetBurstShrdPtr burst = ChaosMakeSharedPtr<DatasetBurst>(db_sdw());

  if (burst->type == chaos::ControlUnitNodeDefinitionType::DSStorageBurstTypeUndefined) {
    ACULERR_ << CHAOS_FORMAT("The type is mandatory for burst %1%", % data->getJSONString());
    return CDWUniquePtr();
  }
  if (!burst->value.isValid()) {
    ACULERR_ << CHAOS_FORMAT("The value is mandatory for burst %1%", % data->getJSONString());
    return CDWUniquePtr();
  }
  ACULDBG_ << "Enabling burst:" << data->getCompliantJSONString();
  LQueueBurstWriteLock wl = burst_queue.getWriteLockObject();
  burst_queue().push(burst);
  return CDWUniquePtr();
}

void AbstractControlUnit::addTag(const std::string& tname) {
  key_data_storage->setOverrideStorageType(DataServiceNodeDefinitionType::DSStorageTypeHistory);
  *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_STATE)->getValuePtr<bool>() = true;
  attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_TAG)->setStringValue(tname, true, true);
  key_data_storage->addTag(tname);
}
void AbstractControlUnit::removeTag(const std::string& tname) {
  key_data_storage->setOverrideStorageType(DataServiceNodeDefinitionType::DSStorageTypeUndefined);

  key_data_storage->removeTag(tname);
  *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_STATE)->getValuePtr<bool>() = false;
  attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_TAG)->setStringValue("");
  *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_CNT_DOWN)->getValuePtr<int32_t>() = 0;

  attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
}

CDWUniquePtr AbstractControlUnit::_datasetTagManagement(CDWUniquePtr data) {
  CHECK_ASSERTION_THROW_AND_LOG(data.get() != NULL, ACULERR_, -1, "No parameter found");
  if (data->hasKey(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_ADD_LIST)) {
    CHECK_KEY_THROW_AND_LOG(data,
                            ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_ADD_LIST,
                            ACULERR_,
                            -2,
                            CHAOS_FORMAT("The key %1% need to be a vector", % ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_ADD_LIST));
    ChaosStringSet                 ss;
    CMultiTypeDataArrayWrapperSPtr vec = data->getVectorValue(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST);
    for (int idx = 0; idx < vec->size(); idx++) {
      if (vec->isStringElementAtIndex(idx)) {
        ss.insert(vec->getStringElementAtIndex(idx));
      }
    }
    key_data_storage->addTag(ss);
  }

  if (data->hasKey(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST)) {
    CHECK_KEY_THROW_AND_LOG(data,
                            ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST,
                            ACULERR_,
                            -3,
                            CHAOS_FORMAT("The key %1% need to be a vector", % ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST));
    ChaosStringSet                 ss;
    CMultiTypeDataArrayWrapperSPtr vec = data->getVectorValue(ControlUnitNodeDomainAndActionRPC::ACTION_DATASET_TAG_MANAGEMENT_REMOVE_LIST);
    for (int idx = 0; idx < vec->size(); idx++) {
      if (vec->isStringElementAtIndex(idx)) {
        ss.insert(vec->getStringElementAtIndex(idx));
      }
    }
    key_data_storage->removeTag(ss);
  }

  return CDWUniquePtr();
}

/*
     Get the current control unit state
     */
CDWUniquePtr AbstractControlUnit::_getInfo(CDWUniquePtr getStatedParam) {
  CreateNewDataWrapper(stateResult, );
  //set the string representing the type of the control unit
  stateResult->addStringValue(NodeDefinitionKey::NODE_TYPE, NodeType::NODE_TYPE_CONTROL_UNIT);
  stateResult->addStringValue(NodeDefinitionKey::NODE_SUB_TYPE, control_unit_type);
  return stateResult;
}

void AbstractControlUnit::_updateAcquistionTimestamp(uint64_t alternative_ts) {
  *timestamp_acq_cached_value->getValuePtr<uint64_t>() = alternative_ts / 1000;
  if (use_custom_high_resolution_timestamp) {
    *timestamp_hw_acq_cached_value->getValuePtr<uint64_t>() = alternative_ts;
  }
}

void AbstractControlUnit::useCustomHigResolutionTimestamp(bool _use_custom_high_resolution_timestamp) {
  use_custom_high_resolution_timestamp = _use_custom_high_resolution_timestamp;
}

void AbstractControlUnit::setHigResolutionAcquistionTimestamp(uint64_t high_resolution_timestamp) {
    if (timestamp_hw_acq_cached_value) {

      *timestamp_hw_acq_cached_value->getValuePtr<uint64_t>() = high_resolution_timestamp;
    }
  
}
  void AbstractControlUnit::setOutputTimestamp(uint64_t timestamp){
    *timestamp_acq_cached_value->getValuePtr<uint64_t>()=timestamp;
  }

void AbstractControlUnit::_updateRunScheduleDelay(uint64_t new_scehdule_delay) {
  if (*thread_schedule_daly_cached_value->getValuePtr<uint64_t>() == new_scehdule_delay) return;
  //we need to update the value
  *thread_schedule_daly_cached_value->getValuePtr<uint64_t>() = new_scehdule_delay;
  thread_schedule_daly_cached_value->markAsChanged();
}

//!timer for update push metric
void AbstractControlUnit::_updatePushRateMetric() {
  uint64_t rate_acq_ts      = TimingUtil::getTimeStamp();
  double   time_offset      = (double(rate_acq_ts - last_push_rate_grap_ts)) / 1000.0;                    //time in seconds
  double   output_ds_rate   = (time_offset > 0) ? push_dataset_counter / time_offset : 0;                 //rate in seconds
  int32_t  output_size_rate = (push_dataset_counter > 0) ? push_dataset_size / push_dataset_counter : 0;  //rate in seconds
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_RATE,
                                                  output_ds_rate);

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_ALARM_LEVEL,
                                                  std::max((map_variable_catalog.count(StateVariableTypeAlarmCU)?map_variable_catalog[StateVariableTypeAlarmCU].maxLevel():0), (map_variable_catalog.count(StateVariableTypeAlarmDEV)?map_variable_catalog[StateVariableTypeAlarmDEV].maxLevel():0)));

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_ALARM_MASKED,
                                                  (map_variable_catalog.count(StateVariableTypeAlarmCU)?map_variable_catalog[StateVariableTypeAlarmCU].countMask():0)+ (map_variable_catalog.count(StateVariableTypeAlarmDEV)?map_variable_catalog[StateVariableTypeAlarmDEV].countMask():0));

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_TSOFF,
                                                  chaos::common::utility::TimingUtil::mds_calibration_offset);

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_SIZE,
                                                  output_size_rate);

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_ERROR,
                                                  (int32_t)push_errors);

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_DATASET_PUSH_LOST,
                                                  (int32_t)packet_lost);

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  ControlUnitHealtDefinitionValue::CU_HEALT_OUTPUT_TOT_PUSH_KSIZE,
                                                  (int32_t)(push_tot_size / 1024));
  setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_lost", chaos::common::alarm::MultiSeverityAlarmLevelClear);
  setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_send_error", chaos::common::alarm::MultiSeverityAlarmLevelClear);

  //keep track of acquire timestamp
  last_push_rate_grap_ts = rate_acq_ts;
  //reset pushe count
  push_dataset_counter = 0;
  push_dataset_size    = 0;
}

//!put abstract control unit state machine in recoverable error
void AbstractControlUnit::_goInRecoverableError(chaos::CException recoverable_exception) {
  //change state machine
  if (SWEService::goInRecoverableError(this, recoverable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
    //update healt the status to report recoverable error
  }
}

//!put abstract control unit state machine in fatal error
void AbstractControlUnit::_goInFatalError(chaos::CException recoverable_exception) {
  //change state machine
  if (SWEService::goInFatalError(this, recoverable_exception, "AbstractControlUnit", __PRETTY_FUNCTION__)) {
    fatalErrorHandler(recoverable_exception);
  }
}

void AbstractControlUnit::goInFatalError(std::string msg, int err, std::string domain) {
  CFatalException ex(err, msg, domain);
  _goInFatalError(ex);
  std::stringstream ss;
  ss << "Fatal Error:\"" << msg << "\" domain:\"" << domain << "\" errcode:" << err;
  metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelError, ss.str());
}
void AbstractControlUnit::fatalErrorHandler(const chaos::CException& ex) {
  ACULERR_ << ex.errorMessage;
}

void AbstractControlUnit::_completeDatasetAttribute() {
  // alarms

  //add global alarm checn
  /*
         DatasetDB::addAttributeToDataSet(stateVariableEnumToName(StateVariableTypeAlarmCU),
         "Activated when some warning has been issued",
         DataType::TYPE_INT32,
         DataType::Output);
         DatasetDB::addAttributeToDataSet(stateVariableEnumToName(StateVariableTypeAlarmDEV),
         "Activated when some alarm has been issued",
         DataType::TYPE_INT32,
         DataType::Output);
         */
}

void AbstractControlUnit::_setBypassState(bool bypass_stage,
                                          bool high_priority) {
  DrvMsg cmd;
  cmd.opcode = bypass_stage ? OpcodeType::OP_SET_BYPASS : OpcodeType::OP_CLEAR_BYPASS;
  //broadcast bypass to all driver instances allocated by control unit
  for (VInstantitedDriverIterator it  = accessor_instances.begin(),
                                  end = accessor_instances.end();
       it != end;
       it++) {
   // (*it)->send(&cmd, chaos::common::constants::CUTimersTimeoutinMSec);
  (*it)->setBypass(bypass_stage);
  }
   ACULDBG_ << "BYPASS COMMAND:"<<bypass_stage;
  setBypassFlag(bypass_stage);

  //update dateset
}
//!handler calledfor restor a control unit to a determinate point
bool AbstractControlUnit::unitRestoreToSnapshot(AbstractSharedDomainCache* const snapshot_cache) {
  return true;
}
void AbstractControlUnit::setState(const std::string& state,bool update){
  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  state,
                                                  update);
}

chaos::common::data::CDWUniquePtr AbstractControlUnit::_unitPerformCalibration(chaos::common::data::CDWUniquePtr data) {
  chaos::common::data::CDWUniquePtr ret;

  HealtManager::getInstance()->addNodeMetricValue(control_unit_id,
                                                  NodeHealtDefinitionKey::NODE_HEALT_STATUS,
                                                  NodeHealtDefinitionValue::NODE_HEALT_STATUS_CALIBRATE,
                                                  true);

  ret = unitPerformCalibration(MOVE(data));
  return ret;
}

chaos::common::data::CDWUniquePtr AbstractControlUnit::unitPerformCalibration(chaos::common::data::CDWUniquePtr data) {
  ACULDBG_ << "Empty calibration:" << (data.get() ? data->getJSONString() : "");
  return CDWUniquePtr();
}

//! this andler is called befor the input attribute will be updated
bool AbstractControlUnit::unitInputAttributePreChangeHandler(chaos::common::data::CDWUniquePtr& pp) {
  return true;
}

//! attribute change handler
/*!
     the handle is fired after the input attribute cache as been update triggere
     by the rpc request for attribute change.
     */
void AbstractControlUnit::unitInputAttributeChangedHandler() {}

template <typename t>
void checkForRange(t v, RangeValueInfo& attributeInfo, const std::string& attr_name, const std::string& id) {
  t                 max, min;
  std::size_t       mini = attributeInfo.minRange.find("0x");
  std::size_t       maxi = attributeInfo.maxRange.find("0x");
  std::stringstream maxs, mins;

  if (maxi != std::string::npos) {
    maxs << std::hex << attributeInfo.maxRange;
  } else {
    maxs << attributeInfo.maxRange;
  }
  if (mini != std::string::npos) {
    mins << std::hex << attributeInfo.minRange;

  } else {
    mins << attributeInfo.minRange;
  }

  maxs >> max;
  mins >> min;
  /*
    try{
        min=boost::lexical_cast<t>(mins);
    }catch(std::exception &e){
        throw MetadataLoggingCException(id, -1, boost::str(boost::format("Invalid casting MIN value (%1%) for attribute %2% value %3%" ) % mins % attr_name % v).c_str(), __PRETTY_FUNCTION__);
    }
    try{

    max=boost::lexical_cast<t>(maxs);
    } catch(std::exception &e){
        throw MetadataLoggingCException(id, -1, boost::str(boost::format("Invalid casting MAX value (%1%) for attribute %2% value %3%" ) % maxs % attr_name % v).c_str(), __PRETTY_FUNCTION__);
    }
  */
  if (((attributeInfo.minRange.size() > 0) && (v < min)) || ((attributeInfo.maxRange.size() > 0) && (v > max))) {
    throw MetadataLoggingCException(id, -1, boost::str(boost::format("Invalid value (%1%) [Min:%2%-%3% Max:%4%-%5%] for attribute %6%") % v % min % attributeInfo.minRange % max % attributeInfo.maxRange % attr_name).c_str(), __PRETTY_FUNCTION__);
  }
}
#define CHECK_FOR_RANGE_VALUE(t, v, attr_name) \
  checkForRange<t>(v, attributeInfo, attr_name, getCUID());

/*
#define CHECK_FOR_RANGE_VALUE(t, v, attr_name)                                                                            \
t max, min;                                                                                                             \
max = strtoll(attributeInfo.maxRange.c_str(), 0, 0);                                                                  \
min = strtoll(attributeInfo.minRange.c_str(), 0, 0);\
if (((attributeInfo.minRange.size()>0)&&(v < min)) || ((attributeInfo.maxRange.size()>0)&&(v > max))) throw MetadataLoggingCException(getCUID(), -1, boost::str(boost::format("Invalid value (%1%) [Min:%2%-%3% Max:%4%-%5%] for attribute %6%") % v % min  % attributeInfo.minRange % max % attributeInfo.maxRange % attr_name).c_str(), __PRETTY_FUNCTION__);
  */
#define CHECK_FOR_STRING_RANGE_VALUE(v, attr_name)                                                                                                                                                                                                                                                      \
  if (attributeInfo.minRange.size() && v < attributeInfo.minRange) throw MetadataLoggingCException(getCUID(), -1, boost::str(boost::format("Invalid value (%1%) [max:%2% Min:%3%] for attribute %4%") % v % attr_name % attributeInfo.minRange % attributeInfo.maxRange).c_str(), __PRETTY_FUNCTION__); \
  if (attributeInfo.maxRange.size() && v > attributeInfo.maxRange) throw MetadataLoggingCException(getCUID(), -1, boost::str(boost::format("Invalid value (%1%) [max:%2% Min:%3%] for attribute %4%") % v % attr_name % attributeInfo.minRange % attributeInfo.maxRange).c_str(), __PRETTY_FUNCTION__);

CDWUniquePtr AbstractControlUnit::setDatasetAttribute(CDWUniquePtr dataset_attribute_values) {
  CHAOS_ASSERT(dataset_attribute_values.get())
  ChaosSharedPtr<SharedCacheLockDomain> w_lock = attribute_value_shared_cache->getLockOnDomain(DOMAIN_INPUT, true);
  w_lock->lock();

  RangeValueInfo           attributeInfo;
  std::vector<std::string> in_attribute_name;

  try {
    //call pre handler
    if (unitInputAttributePreChangeHandler(dataset_attribute_values)) {
      //first call attribute handler
      dataset_attribute_manager.executeHandlers(dataset_attribute_values.get());
    }

    //get all input attribute name for input and bidirectional directions
    getDatasetAttributesName(DataType::Input, in_attribute_name);
    getDatasetAttributesName(DataType::Bidirectional, in_attribute_name);
    attribute_value_shared_cache->getAttributeNames(DOMAIN_CUSTOM, in_attribute_name);
    if (dataset_attribute_values->hasKey(NodeDefinitionKey::NODE_UNIQUE_ID)) {
      //get the contrl unit id
      std::string node_id = dataset_attribute_values->getStringValue(NodeDefinitionKey::NODE_UNIQUE_ID);
      //compare the message device id and the local
      for (std::vector<std::string>::iterator iter = in_attribute_name.begin();
           iter != in_attribute_name.end();
           iter++) {
        //execute attribute handler
        const char* attr_name = iter->c_str();

        //check if the attribute name is present
        if (dataset_attribute_values->hasKey(attr_name)) {
          //check if attribute has been accepted
          if (dataset_attribute_manager.getHandlerResult(*iter) == false) continue;

          AttributeValue* attribute_cache_value = attribute_value_shared_cache->getAttributeValue(DOMAIN_INPUT, attr_name);

          //get attribute info
          getAttributeRangeValueInfo(*iter, attributeInfo);

          //call handler
          switch (attribute_cache_value->type) {
            case DataType::TYPE_BOOLEAN: {
              bool bv = dataset_attribute_values->getValue<bool>(attr_name);  //dataset_attribute_values->getBoolValue(attr_name);
              attribute_cache_value->setValue(&bv, sizeof(bool));
              break;
            }
            case DataType::TYPE_INT32: {
              int32_t i32v = dataset_attribute_values->getValue<int32_t>(attr_name);  //dataset_attribute_values->getInt32Value(attr_name);
              CHECK_FOR_RANGE_VALUE(int32_t, i32v, attr_name)
              attribute_cache_value->setValue(&i32v, sizeof(int32_t));
              break;
            }
            case DataType::TYPE_INT64: {
              int64_t i64v = dataset_attribute_values->getValue<int64_t>(attr_name);  //dataset_attribute_values->getInt64Value(attr_name);
              CHECK_FOR_RANGE_VALUE(int64_t, i64v, attr_name)
              attribute_cache_value->setValue(&i64v, sizeof(int64_t));
              break;
            }
            case DataType::TYPE_DOUBLE: {
              double dv = dataset_attribute_values->getValue<double>(attr_name);  //dataset_attribute_values->getDoubleValue(attr_name);
              CHECK_FOR_RANGE_VALUE(double, dv, attr_name)
              attribute_cache_value->setValue(&dv, sizeof(double));
              break;
            }
            case DataType::TYPE_FLOAT: {
              float dv = dataset_attribute_values->getValue<float>(attr_name);  //dataset_attribute_values->getDoubleValue(attr_name);
              CHECK_FOR_RANGE_VALUE(float, dv, attr_name)
              attribute_cache_value->setValue(&dv, sizeof(float));
              break;
            }

            case DataType::TYPE_JSON: {
              // ChaosUniquePtr<CDataWrapper> str = dataset_attribute_values->getCSDataValue(attr_name);
              CDataWrapper cw;
              try {
                /* if (str.get()) {
                                        attribute_cache_value->setValue(*(str.get()));
                                    }*/
                cw.setSerializedJsonData(dataset_attribute_values->getStringValue(attr_name).c_str());
                attribute_cache_value->setValue(cw);

              } catch (...) {
                throw MetadataLoggingCException(getCUID(), -1, boost::str(boost::format("Invalid Json format ")).c_str(), __PRETTY_FUNCTION__);
              }
              break;
            }
            case DataType::TYPE_STRING: {
              std::string str = dataset_attribute_values->getStringValue(attr_name);
              //CHECK_FOR_STRING_RANGE_VALUE(str, attr_name)
              attribute_cache_value->setValue(str.c_str(), (uint32_t)str.size() + 1);
              break;
            }
            case DataType::TYPE_BYTEARRAY: {
              chaos::common::data::CDBufferUniquePtr bin      = dataset_attribute_values->getBinaryValueAsCDataBuffer(attr_name);
              uint32_t                               bin_size = bin->getBufferSize();
              const char*                            binv     = bin->getBuffer();  //dataset_attribute_values->getBinaryValue(attr_name, bin_size);
              attribute_cache_value->setValue(binv, bin_size);
              break;
            }
            default:
              break;
          }
        }
      }
      w_lock->unlock();

      //push the input attribute dataset
      pushInputDataset();
    }

    //inform the subclass for the change
    unitInputAttributeChangedHandler();
  } catch (CException& ex) {
    //inform the subclass for the change
    unitInputAttributeChangedHandler();
    throw;
  }
  return CDWUniquePtr();
}

/*
     Update the configuration for all descendant tree in the Control Unit class structure
     */
CDWUniquePtr AbstractControlUnit::updateConfiguration(CDWUniquePtr update_pack) {
  //check to see if the device can ben initialized
  if (SWEService::getServiceState() != chaos::CUStateKey::INIT &&
      SWEService::getServiceState() != chaos::CUStateKey::START) {
    ACULERR_ << "device:" << DatasetDB::getDeviceID() << " not initialized";
    throw MetadataLoggingCException(getCUID(), -3, "Device Not Initilized", __PRETTY_FUNCTION__);
  }

  PropertyGroupVectorSDWrapper pg_sdw;
  pg_sdw.serialization_key = "property";
  pg_sdw.deserialize(update_pack.get());
  ACULDBG_ << "properties " << DatasetDB::getDeviceID() << " :" << pg_sdw.serialize()->getJSONString();
  key_data_storage->updateConfiguration(update_pack.get());

  //update the property
  PropertyCollector::applyValue(pg_sdw());
  //mark all cache as changed
  attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();

  pushSystemDataset();
  return CDWUniquePtr();
}

bool AbstractControlUnit::propertyChangeHandler(const std::string&  group_name,
                                                const std::string&  property_name,
                                                const CDataVariant& property_value) {
  ACULDBG_ << CHAOS_FORMAT("Update property request for %1%[%2%] with value %3%", % property_name % group_name % property_value.asString());
  return true;
}

void AbstractControlUnit::propertyUpdatedHandler(const std::string&  group_name,
                                                 const std::string&  property_name,
                                                 const CDataVariant& old_value,
                                                 const CDataVariant& new_value) {
  if (group_name.compare("property_abstract_control_unit") == 0) {
    //update property on driver
    key_data_storage->updateConfiguration(property_name, new_value);
    //reflect modification on dataset
    if (property_name.compare(ControlUnitDatapackSystemKey::BYPASS_STATE) == 0) {
      _setBypassState(new_value.asBool());
    } else if (property_name.compare(DataServiceNodeDefinitionKey::DS_STORAGE_TYPE) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, DataServiceNodeDefinitionKey::DS_STORAGE_TYPE)->getValuePtr<int32_t>() = new_value.asInt32();
    } else if (property_name.compare(DataServiceNodeDefinitionKey::DS_STORAGE_LIVE_TIME) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, DataServiceNodeDefinitionKey::DS_STORAGE_LIVE_TIME)->getValuePtr<uint64_t>() = new_value.asUInt64();
    } else if (property_name.compare(DataServiceNodeDefinitionKey::DS_STORAGE_HISTORY_TIME) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, DataServiceNodeDefinitionKey::DS_STORAGE_HISTORY_TIME)->getValuePtr<uint64_t>() = new_value.asUInt64();
    }  else if (property_name.compare(DataServiceNodeDefinitionKey::DS_STORAGE_LOG_TIME) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, DataServiceNodeDefinitionKey::DS_STORAGE_LOG_TIME)->getValuePtr<uint64_t>() = new_value.asUInt64();
    } else if (property_name.compare(DataServiceNodeDefinitionKey::DS_UPDATE_ANYWAY) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, DataServiceNodeDefinitionKey::DS_UPDATE_ANYWAY)->getValuePtr<int32_t>() = new_value.asInt32();
      ds_update_anyway                                                                                                                        = new_value.asInt32();
    } else if (property_name.compare(ControlUnitDatapackSystemKey::CU_LOG_MAX_MS) == 0) {
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::CU_LOG_MAX_MS)->getValuePtr<int32_t>() = new_value.asInt32();
      log_maxupdate_ms                                                                                                                     = new_value.asInt32();
    }
  }
}

//! return the accessor by an index
/*
     The index parameter correspond to the order that the driver infromation are
     added by the unit implementation into the function AbstractControlUnit::unitDefineDriver.
     */
DriverAccessor* AbstractControlUnit::getAccessoInstanceByIndex(int idx) {
  if (idx >= accessor_instances.size()) {
    ACULERR_ << "  driver instance index:" << idx << "not present,# available drivers:" << accessor_instances.size();

    return NULL;
  }
  return accessor_instances[idx];
}

int AbstractControlUnit::pushOutputDataset() {
  int                                   err                    = 0;
  AttributeCache&                       output_attribute_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT);
  ChaosSharedPtr<SharedCacheLockDomain> r_lock                 = attribute_value_shared_cache->getLockOnDomain(DOMAIN_OUTPUT, false);
  r_lock->lock();
  uint64_t tscor = *timestamp_acq_cached_value->getValuePtr<uint64_t>();//TimingUtil::getTimeStamp();//TimingUtil::getTimeCorStamp();
  if ((tscor - last_push) < ds_update_anyway) {
    //check if something as changed
    if (!output_attribute_cache.hasChanged()) {
      //ACULDBG_ << " Nothing changed in "<<(tscor - last_push)<<" ms";

      return err;
    }
  }
  if((aligned_ex_done==false)&&(push_dataset_counter==1)){
    dsInitSetFromReadout();
    aligned_ex_done=true;
  }
  last_push                           = tscor;
  CDWShrdPtr output_attribute_dataset = key_data_storage->getNewDataPackForDomain(KeyDataStorageDomainOutput);
  if (!output_attribute_dataset.get()) {
    ACULERR_ << " Cannot allocate packet.. err:" << err;
    return err;
  }
  if ((busy == false)&&(bypass==false)) {
    checkAlarms();
  } /*else {
    ACULDBG_<<"Check disable because busy:"<<busy<<" bypass:"<<bypass;
  }*/

  output_attribute_dataset->addInt64Value(ControlUnitDatapackCommonKey::RUN_ID, run_id);
 // output_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP,tscor);
  //output_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP, *timestamp_hw_acq_cached_value->getValuePtr<uint64_t>());
  //ACULDBG_<<"TIME DIFF:"<<(tscor-(*timestamp_hw_acq_cached_value->getValuePtr<uint64_t>()/1000));

  //add all other output channel
  for (int idx = 0;
       idx < ((int)cache_output_attribute_vector.size());  //the device id and timestamp in added out of this list
       idx++) {
    //
    AttributeValue* value_set = cache_output_attribute_vector[idx];
    assert(value_set);
    switch (value_set->type) {
      case DataType::TYPE_BOOLEAN:
        output_attribute_dataset->addBoolValue(value_set->name, *value_set->getValuePtr<bool>());
        break;
      case DataType::TYPE_INT32:
        //ACULDBG_ << value_set->name << " <="<<*value_set->getValuePtr<int32_t>();

        output_attribute_dataset->addInt32Value(value_set->name, *value_set->getValuePtr<int32_t>());
        break;
      case DataType::TYPE_INT64:
        output_attribute_dataset->addInt64Value(value_set->name, *value_set->getValuePtr<int64_t>());
        break;
      case DataType::TYPE_UINT64:
        output_attribute_dataset->addUInt64Value(value_set->name, *value_set->getValuePtr<uint64_t>());
        break;
      case DataType::TYPE_DOUBLE:
        output_attribute_dataset->addDoubleValue(value_set->name, *value_set->getValuePtr<double>());
        break;
       case DataType::TYPE_FLOAT:
        output_attribute_dataset->addDoubleValue(value_set->name, *value_set->getValuePtr<float>());
        break;
      case DataType::TYPE_JSON: {
        try {
          output_attribute_dataset->addCSDataValue(value_set->name, *value_set->getValuePtr<CDataWrapper>());
        } catch (...) {
          throw MetadataLoggingCException(getCUID(), -101, boost::str(boost::format("Invalid Json format for attribute '%1%' :'%2%") % value_set->name % value_set->getValuePtr<const char>()).c_str(), __PRETTY_FUNCTION__);
        }
        break;
      }
      case DataType::TYPE_STRING:
        //DEBUG_CODE(ACULDBG_ << value_set->name<<"-"<<value_set->getValuePtr<const char>();)
        output_attribute_dataset->addStringValue(value_set->name, value_set->getValuePtr<const char>());
        break;
      case DataType::TYPE_BYTEARRAY:
        if (value_set->size) {
          if ((value_set->sub_type.size() == 1) && (value_set->sub_type[0] != DataType::SUB_TYPE_MIME)) {
            output_attribute_dataset->addBinaryValue(value_set->name, value_set->sub_type[0], value_set->getValuePtr<char>(), value_set->size);
          } else {
            output_attribute_dataset->addBinaryValue(value_set->name, value_set->getValuePtr<char>(), value_set->size);
          }
        }
        break;
      default:
        break;
    }
  }
  //manage the burst information
  manageBurstQueue();
  //now we nede to push the outputdataset
  int psiz = output_attribute_dataset->getBSONRawSize();
  push_dataset_size += psiz;
  push_tot_size += psiz;
  int retry = 10;
  do {
    err = key_data_storage->pushDataSet(data_manager::KeyDataStorageDomainOutput, output_attribute_dataset);
    if (err != 0) {
      push_errors++;
      ERR << push_errors << "] ERROR pushing runid:" << run_id;
      setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_send_error", chaos::common::alarm::MultiSeverityAlarmLevelWarning);

      usleep(10000);
    }
  } while ((err != 0) && (retry--));
  if (retry < 0) {
    packet_lost++;
    setStateVariableSeverity(StateVariableTypeAlarmCU, "packet_lost", (common::alarm::MultiSeverityAlarmLevel)packet_lost);

    // inform higher levels...
  }
  //update counter
  push_dataset_counter++;

  //reset chagned attribute into output dataset
  if (!err) {
    output_attribute_cache.resetChangedIndex();
  }
  return err;
}

//push system dataset
int AbstractControlUnit::pushInputDataset() {
  int             err                   = 0;
  AttributeCache& input_attribute_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT);
  if (!input_attribute_cache.hasChanged()) return err;
  //get the cdatawrapper for the pack
 // int64_t    cur_us                  = TimingUtil::getTimeStampInMicroseconds();
  CDWShrdPtr input_attribute_dataset = key_data_storage->getNewDataPackForDomain(KeyDataStorageDomainInput);
  if (input_attribute_dataset.get()) {
    input_attribute_dataset->addInt64Value(ControlUnitDatapackCommonKey::RUN_ID, run_id);
    //input dataset timestamp is added only when pushed on cache
   // input_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP, TimingUtil::getTimeCorStamp());
   input_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP, TimingUtil::getTimeStamp());
   
   // input_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP, cur_us);
    //fill the dataset
    fillCDatawrapperWithCachedValue(cache_input_attribute_vector, *input_attribute_dataset);

    //push out the system dataset
    err = key_data_storage->pushDataSet(data_manager::KeyDataStorageDomainInput, MOVE(input_attribute_dataset));
    if (!err) {
      input_attribute_cache.resetChangedIndex();
    }
  } else {
    ACULERR_ << " Cannot allocate packet.. err:" << err;
  }
  return err;
}

//push system dataset
int AbstractControlUnit::pushCustomDataset() {
  int             err                    = 0;
  AttributeCache& custom_attribute_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_CUSTOM);
  if (!custom_attribute_cache.hasChanged()) return err;
  //get the cdatawrapper for the pack
 // int64_t cur_us = TimingUtil::getTimeStampInMicroseconds();
  if (key_data_storage.get()) {
    CDWShrdPtr custom_attribute_dataset = key_data_storage->getNewDataPackForDomain(KeyDataStorageDomainCustom);
    if (custom_attribute_dataset.get()) {
      custom_attribute_dataset->addInt64Value(ControlUnitDatapackCommonKey::RUN_ID, run_id);
      //input dataset timestamp is added only when pushed on cache
      custom_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP, TimingUtil::getTimeStamp());
    //  custom_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP, cur_us);

      //fill the dataset
      fillCDatawrapperWithCachedValue(cache_custom_attribute_vector, *custom_attribute_dataset);
      ACULDBG_ << " Push custom:"<<custom_attribute_dataset->getJSONString();

      //push out the system dataset
      err = key_data_storage->pushDataSet(data_manager::KeyDataStorageDomainCustom, MOVE(custom_attribute_dataset));

      if (!err) {
        custom_attribute_cache.resetChangedIndex();
      }
    } else {
      ACULERR_ << " Cannot allocate packet.. err:" << err;
    }
  }
  return err;
}

int AbstractControlUnit::pushSystemDataset() {
  int             err                    = 0;
  AttributeCache& system_attribute_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  uint64_t        tscor                  = TimingUtil::getTimeStamp();
  if ((tscor - last_push) < ds_update_anyway) {
    //check if something as changed
    if (!system_attribute_cache.hasChanged()) {
      return err;
    }
  }
  //get the cdatawrapper for the pack
  int64_t cur_us = TimingUtil::getTimeStampInMicroseconds();
  if (key_data_storage.get() == NULL) {
    ACULERR_ << " key data storage not allocated";

    return -2;
  }
  CDWShrdPtr system_attribute_dataset = key_data_storage->getNewDataPackForDomain(KeyDataStorageDomainSystem);
  if (system_attribute_dataset.get()) {
    system_attribute_dataset->addInt64Value(ControlUnitDatapackCommonKey::RUN_ID, run_id);
    //input dataset timestamp is added only when pushed on cache
    system_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP, tscor);
   // system_attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_HIGH_RESOLUTION_TIMESTAMP, cur_us);
    //fill the dataset
    fillCDatawrapperWithCachedValue(cache_system_attribute_vector, *system_attribute_dataset);
    //ACULDBG_ << system_attribute_dataset->getJSONString();
    //push out the system dataset
    err = key_data_storage->pushDataSet(data_manager::KeyDataStorageDomainSystem, MOVE(system_attribute_dataset));
    if (!err) {
      system_attribute_cache.resetChangedIndex();
    }
  }
  return err;
}

CDWShrdPtr AbstractControlUnit::writeCatalogOnCDataWrapper(AlarmCatalog& catalog,
                                                           int32_t       dataset_type) {
  if (key_data_storage.get() == NULL) {
    return CDWShrdPtr();
  }
  CDWShrdPtr attribute_dataset = key_data_storage->getNewDataPackForDomain((KeyDataStorageDomain)dataset_type);
  if (attribute_dataset) {
    //fill datapack with
    //! the dataaset can be pushed also in other moment
    attribute_dataset->addInt64Value(DataPackCommonKey::DPCK_TIMESTAMP, TimingUtil::getTimeStamp());
    attribute_dataset->addInt64Value(ControlUnitDatapackCommonKey::RUN_ID, run_id);
    //scan all alarm ad create the datapack
    size_t alarm_size = catalog.size();
    for (unsigned int idx = 0;
         idx < alarm_size;
         idx++) {
      AlarmDescription* alarm = catalog.getAlarmByOrderedID(idx);
      attribute_dataset->addInt32Value(alarm->getAlarmName(),
                                       (uint32_t)alarm->getCurrentSeverityCode());

      if (alarm->getMask() != 0xFF) {
        // ACULDBG_ << " MASK:"<<(uint32_t)alarm->getMask()<<" JSON:"<<attribute_dataset->getCompliantJSONString();

        attribute_dataset->addInt32Value(alarm->getAlarmName() + "_MASK",
                                         (uint32_t)alarm->getMask());
      }
    }
  }
  return attribute_dataset;
}

int AbstractControlUnit::pushDevAlarmDataset() {
  GET_CAT_OR_EXIT(StateVariableTypeAlarmDEV, 0);
  int err = 0;

  CDWShrdPtr attribute_dataset = writeCatalogOnCDataWrapper(catalog,
                                                            DataPackCommonKey::DPCK_DATASET_TYPE_DEV_ALARM);
  if (attribute_dataset && key_data_storage.get()) {
    //push out the system dataset
    err = key_data_storage->pushDataSet(KeyDataStorageDomainDevAlarm, MOVE(attribute_dataset));
  }
  return err;
}

int AbstractControlUnit::pushCUAlarmDataset() {
  GET_CAT_OR_EXIT(StateVariableTypeAlarmCU, 0);
  int        err               = 0;
  CDWShrdPtr attribute_dataset = writeCatalogOnCDataWrapper(catalog,
                                                            DataPackCommonKey::DPCK_DATASET_TYPE_CU_ALARM);
  if (attribute_dataset && key_data_storage.get()) {
    //push out the system dataset
    err = key_data_storage->pushDataSet(KeyDataStorageDomainCUAlarm, MOVE(attribute_dataset));
  }
  return err;
}

void AbstractControlUnit::manageBurstQueue() {
  static int64_t old_ts = 0;
  if (!current_burst.get()) {
    DatasetBurstShrdPtr next_burst;
    LQueueBurstReadLock wl = burst_queue.getReadLockObject();
    if (!burst_queue().empty()) {
      next_burst = burst_queue().front();
      burst_queue().pop();
    }
    wl->unlock();

    if (next_burst.get()) {
      switch (next_burst->type) {
        case chaos::ControlUnitNodeDefinitionType::DSStorageBurstTypeNPush:
          current_burst.reset(new PushStorageBurst(next_burst));
          break;
        case chaos::ControlUnitNodeDefinitionType::DSStorageBurstTypeMSec:
          current_burst.reset(new MSecStorageBurst(next_burst));
          break;
        default:
          break;
      }
      old_ts = 0;
      //set the tag for burst
      ACULDBG_ << "======= Start Burst tag:'" << current_burst->dataset_burst->tag << "' info:"<<current_burst->dataset_burst->loginfo<<" =======";
      key_data_storage->addTag(current_burst->dataset_burst->tag);
      key_data_storage->setTimingConfigurationBehaviour(false);
      key_data_storage->setOverrideStorageType(DataServiceNodeDefinitionType::DSStorageTypeHistory);
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_STATE)->getValuePtr<bool>() = true;
      attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_TAG)->setStringValue(current_burst->dataset_burst->tag, true, true);
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_CNT_DOWN)->getValuePtr<int32_t>() = current_burst->remaining();

      attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
      if(current_burst->dataset_burst->loginfo!=""){
        std::stringstream ss;
        ss<<"Start Tagging '"<<current_burst->dataset_burst->tag<<"'\n"<<current_burst->dataset_burst->loginfo;
        
        metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,ss.str());
      }
      pushSystemDataset();
      // mark also setpoints
      attribute_value_shared_cache->getSharedDomain(DOMAIN_INPUT).markAllAsChanged();
      pushInputDataset();
      
    }
  } else {
    int64_t tim = *timestamp_acq_cached_value->getValuePtr<int64_t>();
    if (!current_burst->active(tim)) {
      //remove the tag for the burst
      ACULDBG_ << "======= End Burst tag:'" << current_burst->dataset_burst->tag << "' =======";
      if(current_burst->dataset_burst->loginfo!=""){
        std::stringstream ss;
        ss<<"End Tagging '"<<current_burst->dataset_burst->tag<<"'";
        metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,ss.str());
      }
      key_data_storage->removeTag(current_burst->dataset_burst->tag);
      key_data_storage->setTimingConfigurationBehaviour(true);
      key_data_storage->setOverrideStorageType(DataServiceNodeDefinitionType::DSStorageTypeUndefined);
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_STATE)->getValuePtr<bool>() = false;
      attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_TAG)->setStringValue("");
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_CNT_DOWN)->getValuePtr<int32_t>() = 0;

      attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
      //we need to remove it
      current_burst.reset();
      pushSystemDataset();
    } else {
      int32_t rem                                                                                                                           = current_burst->remaining();
      *attribute_value_shared_cache->getAttributeValue(DOMAIN_SYSTEM, ControlUnitDatapackSystemKey::BURST_CNT_DOWN)->getValuePtr<int32_t>() = rem;
      if ((tim - old_ts > chaos::common::constants::CUTimersTimeoutinMSec) == 0) {
        ACULDBG_ << "======= Active Burst tag:'" << current_burst->dataset_burst->tag << "' remaining:" << rem << " =======";

        attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
        pushSystemDataset();
      }
      old_ts = tim;
    }
  }
}

void AbstractControlUnit::fillCDatawrapperWithCachedValue(std::vector<AttributeValue*>& cached_attributes, CDataWrapper& dataset) {
  for (std::vector<AttributeValue*>::iterator it = cached_attributes.begin();
       it != cached_attributes.end();
       it++) {
    (*it)->writeToCDataWrapper(dataset);
  }
}

//!timer for update push metric
void AbstractControlUnit::timeout() {
  //update push metric
  _updatePushRateMetric();
}

bool AbstractControlUnit::isInputAttributeChangeAuthorizedByHandler(const std::string& attr_name) {
  return dataset_attribute_manager.getHandlerResult(attr_name);
}

void AbstractControlUnit::copyInitConfiguraiton(CDataWrapper& copy) {
  if (init_configuration.get() == NULL) return;

  //copy all key
  init_configuration->copyAllTo(copy);
}

#pragma mark Abstract Control Unit API
void         AbstractControlUnit::addStateVariable(StateVariableType  variable_type,
                                           const std::string& state_variable_name,
                                           const std::string& state_variable_description,
                                           int32_t            max_freq_ms) {
  if (map_variable_catalog.count(variable_type) == 0) {
    //add new catalog
    map_variable_catalog.insert(MapStateVariablePair(variable_type, AlarmCatalog(stateVariableEnumToName(variable_type))));
  } 
  
  AlarmCatalog& catalog = map_variable_catalog[variable_type];
  if(catalog.getAlarmByName(state_variable_name)){
    // already present
    return;
  }
  catalog.addAlarm(new MultiSeverityAlarm(stateVariableEnumToName(variable_type),
                                          state_variable_name,
                                          state_variable_description,
                                          max_freq_ms));
  //add this instance as
  catalog.addAlarmHandler(state_variable_name,
                          this);
}

void AbstractControlUnit::setStateVariableSeverity(StateVariableType                                   variable_type,
                                                   const chaos::common::alarm::MultiSeverityAlarmLevel state_variable_severity) {
  GET_CAT_OR_EXIT(variable_type, )
  catalog.setAllAlarmSeverity(state_variable_severity);

  /*
         AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT);
         output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.max()));
         */
  AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.maxLevel()));
}

bool AbstractControlUnit::setStateVariableSeverity(StateVariableType             variable_type,
                                                   const std::string&            state_variable_name,
                                                   const MultiSeverityAlarmLevel state_variable_severity) {
  GET_CAT_OR_EXIT(variable_type, false)
  AlarmDescription* alarm = catalog.getAlarmByName(state_variable_name);
  if (alarm == NULL) {
    ACULERR_<<"Alarm \""<<state_variable_name<<"\" not found level:"<<(int)state_variable_severity;
    return false;
  }
  alarm->setCurrentSeverity(state_variable_severity);
  //update global alarm output attribute
  /*
         AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT);
         output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.max()));
         */
  AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);

  output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.maxLevel()));

  return true;
}

bool AbstractControlUnit::setStateVariableSeverity(StateVariableType             variable_type,
                                                   const unsigned int            state_variable_ordered_id,
                                                   const MultiSeverityAlarmLevel state_variable_severity) {
  GET_CAT_OR_EXIT(variable_type, false)
  AlarmDescription* alarm = catalog.getAlarmByOrderedID(state_variable_ordered_id);
  if (alarm == NULL) return false;
  alarm->setCurrentSeverity(state_variable_severity);
  //update global alarm output attribute
  /*
         AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_OUTPUT);
         output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.isCatalogClear()==false));
         */
  AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  output_cache.getValueSettingByName(stateVariableEnumToName(variable_type))->setValue(CDataVariant(catalog.isCatalogClear() == false));

  return true;
}

bool AbstractControlUnit::getStateVariableSeverity(StateVariableType        variable_type,
                                                   const std::string&       state_variable_name,
                                                   MultiSeverityAlarmLevel& state_variable_severity) {
  GET_CAT_OR_EXIT(variable_type, false)
  AlarmDescription* alarm = catalog.getAlarmByName(state_variable_name);
  if (alarm == NULL) return false;
  state_variable_severity = static_cast<MultiSeverityAlarmLevel>(alarm->getCurrentSeverityCode());
  return true;
}

bool AbstractControlUnit::getStateVariableSeverity(StateVariableType        variable_type,
                                                   const unsigned int       state_variable_ordered_id,
                                                   MultiSeverityAlarmLevel& state_variable_severity) {
  GET_CAT_OR_EXIT(variable_type, false)
  AlarmDescription* alarm = catalog.getAlarmByOrderedID(state_variable_ordered_id);
  if (alarm == NULL) return false;
  state_variable_severity = static_cast<MultiSeverityAlarmLevel>(alarm->getCurrentSeverityCode());
  return true;
}

void AbstractControlUnit::alarmChanged(const std::string& state_variable_tag,
                                       const std::string& state_variable_name,
                                       const int8_t       state_variable_severity) {
  int variable_type = stateVariableNameToEnum(state_variable_tag);
  if (variable_type == -1) return;

  GET_CAT_OR_EXIT((StateVariableType)variable_type, )
  AlarmDescription*   alarm    = catalog.getAlarmByName(state_variable_name);
  MultiSeverityAlarm* alarm_ms = (MultiSeverityAlarm*)alarm;
  CHAOS_ASSERT(alarm);

  //update alarm log
  if ((log_maxupdate_ms == 0) ||
      (alarm_logging_channel && (log_maxupdate_ms > 0) && ((alarm->getLastUpdateTimestamp() - alarm_ms->last_log_ms) > log_maxupdate_ms))) {
    ACULDBG_ << "Log State "<<state_variable_name<<" severity:"<<(int32_t)state_variable_severity<<" last modified:"<<( alarm->getLastUpdateTimestamp() -alarm_ms->last_log_ms)<<" ms freq:"<<alarm_ms->max_freq_log_ms;

    alarm_logging_channel->logAlarm(getCUID(),
                                    "AbstractControlUnit",
                                    *alarm);
    alarm_ms->last_log_ms = alarm->getLastUpdateTimestamp();
  }

  switch ((StateVariableType)variable_type) {
    case StateVariableTypeAlarmCU:
      //update dataset alarm on cds
      pushCUAlarmDataset();
      break;

    case StateVariableTypeAlarmDEV:
      //update dataset alarm on cds
      pushDevAlarmDataset();
      break;
  }
  AttributeCache& output_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  output_cache.getValueSettingByName(stateVariableEnumToName((StateVariableType)variable_type))->setValue(CDataVariant(catalog.maxLevel()));

  attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM).markAllAsChanged();
  pushSystemDataset();
}

void AbstractControlUnit::setBusyFlag(bool state) {
  if(state!=busy){
    AttributeCache& system_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
    busy                         = state;
    if (system_cache.hasName("busy")) {
      system_cache.getValueSettingByName("busy")->setValue(CDataVariant(state));
    }
    pushSystemDataset();

  }
}
void AbstractControlUnit::setBypassFlag(bool state) {
 
  if(state!=bypass){
     AttributeCache& system_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
      if (system_cache.hasName(ControlUnitDatapackSystemKey::BYPASS_STATE)) {
        system_cache.getValueSettingByName(ControlUnitDatapackSystemKey::BYPASS_STATE)->setValue(CDataVariant(state));
      }
    ACULDBG_ << "Bypass:"<<state <<" old:"<<bypass;
    if(state){
      metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                      "Bypass enabled");
    } else {
      metadataLogging(chaos::common::metadata_logging::StandardLoggingChannel::LogLevelInfo,
                      "Bypass disabled");
    }
    bypass=state;
    pushSystemDataset();

  }
  

}
const bool AbstractControlUnit::getBusyFlag() const {
  AttributeCache& system_cache = attribute_value_shared_cache->getSharedDomain(DOMAIN_SYSTEM);
  if (system_cache.hasName("busy")) {
    return system_cache.getValueSettingByName("busy")->getAsVariant().asBool();
  } else {
    return false;
  }
}

void AbstractControlUnit::metadataLogging(const std::string&                                                      subject,
                                          const chaos::common::metadata_logging::StandardLoggingChannel::LogLevel log_level,
                                          const std::string&                                                      message) {
  if (standard_logging_channel == NULL) return;
  //also messages must be tagged
  std::string tag;
  if (current_burst.get()&&current_burst->dataset_burst.get()) {
    tag =current_burst->dataset_burst->tag;
  }
  standard_logging_channel->logMessage(getCUID(),
                                       subject,
                                       log_level,
                                       message,tag);
  switch (log_level) {
    case StandardLoggingChannel::LogLevelInfo:
      ACULDBG_ << "LOGINFO"
               << " subj:" << subject << " msg:" << message;
      break;
    case StandardLoggingChannel::LogLevelDebug:
      ACULDBG_ << "LOGDEBUG"
               << " subj:" << subject << " msg:" << message;
      break;
    case StandardLoggingChannel::LogLevelWarning:
      ACULWRN_ << "LOGWARN"
               << " subj:" << subject << " msg:" << message;
      break;
    case StandardLoggingChannel::LogLevelError:
      ACULERR_ << "LOGERR"
               << " subj:" << subject << " msg:" << message;
      break;
    case StandardLoggingChannel::LogLevelFatal:
      ACULNOTE_ << "LOGFATAL"
                << " subj:" << subject << " msg:" << message;
      break;
  }
}

void AbstractControlUnit::metadataLogging(const StandardLoggingChannel::LogLevel log_level,
                                          const std::string&                     message) {
  metadataLogging("AbstractControlUnit",
                  log_level,
                  message);
}

void AbstractControlUnit::consumerHandler( chaos::common::message::ele_t& data) {
  incomingMessage(data.key, data.cd);
}
void AbstractControlUnit::updateDataSet(chaos::common::data::CDataWrapper& cd, chaos::DataType::DataSetAttributeIOAttribute io) {
  std::vector<std::string> props;
  cd.getAllKey(props);
  for (std::vector<std::string>::iterator i = props.begin(); i != props.end(); i++) {
    if (cd.isInt32Value(*i)) {
      getAttributeCache()->setOutputAttributeValue(*i, cd.getInt32Value(*i));
    } else if (cd.isDoubleValue(*i)) {
      getAttributeCache()->setOutputAttributeValue(*i, cd.getDoubleValue(*i));
    } else if (cd.isInt64Value(*i)) {
      getAttributeCache()->setOutputAttributeValue(*i, cd.getInt64Value(*i));
    } else if (cd.isBoolValue(*i)) {
      getAttributeCache()->setOutputAttributeValue(*i, cd.getBoolValue(*i));
    } else if (cd.isStringValue(*i)) {
      getAttributeCache()->setOutputAttributeValue(*i, cd.getStringValue(*i));
    } else if (cd.isVectorValue(*i)) {
      ///
      chaos::common::data::CMultiTypeDataArrayWrapperSPtr v   = cd.getVectorValue(*i);
      int                                                 siz = v->size();
      if (siz > 0) {
        if (v->isInt32ElementAtIndex(0)) {
          int32_t arr[siz];
          for (int cnt = 0; cnt < siz; cnt++) {
            arr[cnt] = v->getInt32ElementAtIndex(cnt);
          }
          getAttributeCache()->setOutputAttributeValue(*i, (void*)arr, siz * sizeof(int32_t));
        } else if (v->isBoolElementAtIndex(0)) {
          bool arr[siz];
          for (int cnt = 0; cnt < siz; cnt++) {
            arr[cnt] = v->getBoolElementAtIndex(cnt);
          }
          getAttributeCache()->setOutputAttributeValue(*i, (void*)arr, siz * sizeof(bool));
        } else if (v->isInt64ElementAtIndex(0)) {
          int64_t arr[siz];
          for (int cnt = 0; cnt < siz; cnt++) {
            arr[cnt] = v->getInt64ElementAtIndex(cnt);
          }
          getAttributeCache()->setOutputAttributeValue(*i, (void*)arr, siz * sizeof(int64_t));
        } else if (v->isDoubleElementAtIndex(0)) {
          double arr[siz];
          for (int cnt = 0; cnt < siz; cnt++) {
            arr[cnt] = v->getDoubleElementAtIndex(cnt);
          }
          getAttributeCache()->setOutputAttributeValue(*i, (void*)arr, siz * sizeof(double));
        }
      }
    } else if (cd.isBinaryValue(*i)) {
      uint32_t    size;
      const char* ptr = cd.getBinaryValue(*i, size);
      getAttributeCache()->setOutputAttributeValue(*i, (void*)ptr, size);
    }
  }
}

void AbstractControlUnit::updateDatasetFromDriverProperty() {
  for (int idx = 0;
       idx != accessor_instances.size();
       idx++) {
    chaos::common::data::CDWUniquePtr ret = accessor_instances[idx]->getDrvProperties();
    std::vector<std::string>          props;
    if (ret.get() == NULL) {
      ACULDBG_ << "No driver properties";
      return;
    }

    ret->getAllKey(props);
    for (std::vector<std::string>::iterator i = props.begin(); i != props.end(); i++) {
      if (ret->isCDataWrapperValue(*i)) {
        chaos::common::data::CDWUniquePtr cd = ret->getCSDataValue(*i);
        if (cd->hasKey(PROPERTY_VALUE_KEY) && cd->hasKey(PROPERTY_VALUE_PUB_KEY)) {
          std::string attrname = cd->getStringValue(PROPERTY_VALUE_PUB_KEY);
          if (cd->hasKey(PROPERTY_VALUE_ALARM_LVL_KEY)) {
            setStateVariableSeverity(StateVariableTypeAlarmDEV, attrname, (common::alarm::MultiSeverityAlarmLevel)cd->getInt32Value(PROPERTY_VALUE_ALARM_LVL_KEY));
          }
          DataType::DataSetAttributeIOAttribute type = chaos::DataType::Bidirectional;
          if (cd->hasKey(PROPERTY_VALUE_IO_KEY)) {
            type = (DataType::DataSetAttributeIOAttribute)cd->getInt32Value(PROPERTY_VALUE_IO_KEY);
          }
          if ((type == chaos::DataType::Bidirectional) || (type == chaos::DataType::Output)) {
            chaos::DataType::DataType dstype = cd->getValueType(PROPERTY_VALUE_KEY);

            //getAttributeCache()->setOutputAttributeValue(attrname,cd->getVariantValue(PROPERTY_VALUE_KEY));
            if (dstype == chaos::DataType::TYPE_DOUBLE) {
              getAttributeCache()->setOutputAttributeValue(attrname, cd->getDoubleValue(PROPERTY_VALUE_KEY));

            } else if (dstype == chaos::DataType::TYPE_INT32) {
              getAttributeCache()->setOutputAttributeValue(attrname, cd->getInt32Value(PROPERTY_VALUE_KEY));

            } else if (dstype == chaos::DataType::TYPE_INT64) {
              getAttributeCache()->setOutputAttributeValue(attrname, cd->getInt64Value(PROPERTY_VALUE_KEY));

            } else if (dstype == chaos::DataType::TYPE_BOOLEAN) {
              getAttributeCache()->setOutputAttributeValue(attrname, cd->getBoolValue(PROPERTY_VALUE_KEY));

            } else if (dstype == chaos::DataType::TYPE_STRING) {
              getAttributeCache()->setOutputAttributeValue(attrname, cd->getStringValue(PROPERTY_VALUE_KEY));
            } else if (dstype == chaos::DataType::TYPE_BYTEARRAY) {
              uint32_t    size;
              const char* ptr = cd->getBinaryValue(PROPERTY_VALUE_KEY, size);
              getAttributeCache()->setOutputAttributeValue(attrname, (void*)ptr, size);
            }
          }
        }
      }
    }
  }
 // getAttributeCache()->setInputDomainAsChanged();
  getAttributeCache()->setOutputDomainAsChanged();
}

bool AbstractControlUnit::setDrvProp(const std::string& name, const bool value, uint32_t size) {
  bool bidir = getAttributeCache()->exist(DOMAIN_OUTPUT, name);
  return _setDrvProp(name, value, size, bidir);
}
bool AbstractControlUnit::setDrvProp(const std::string& name, const int32_t value, uint32_t size) {
  bool bidir = getAttributeCache()->exist(DOMAIN_OUTPUT, name);

  return _setDrvProp(name, value, size, bidir);
}
bool AbstractControlUnit::setDrvProp(const std::string& name, const int64_t value, uint32_t size) {
  bool bidir = getAttributeCache()->exist(DOMAIN_OUTPUT, name);

  return _setDrvProp(name, value, size, bidir);
}
bool AbstractControlUnit::setDrvProp(const std::string& name, const double value, uint32_t size) {
  bool bidir = getAttributeCache()->exist(DOMAIN_OUTPUT, name);

  return _setDrvProp(name, value, size, bidir);
}
bool AbstractControlUnit::setDrvProp(const std::string& name, std::string value, uint32_t size) {
  bool bidir = getAttributeCache()->exist(DOMAIN_OUTPUT, name);

  return _setDrvProp(name, value, size, bidir);
}

void AbstractControlUnit::addAttributesToDataSet(chaos::common::data::CDataWrapper& cd, chaos::DataType::DataSetAttributeIOAttribute io) {
  std::vector<std::string> props;
  cd.getAllKey(props);
  for (std::vector<std::string>::iterator i = props.begin(); i != props.end(); i++) {
    if (cd.isVector(*i)) {
      chaos::common::data::CMultiTypeDataArrayWrapperSPtr v   = cd.getVectorValue(*i);
      int                                                 siz = v->size();
      if (siz > 0) {
        std::stringstream ss;
        ss << siz;
        if (v->isInt32ElementAtIndex(0)) {
          addBinaryAttributeAsSubtypeToDataSet(*i, "int32 vect:" + ss.str(), chaos::DataType::SUB_TYPE_INT32, siz * sizeof(int32_t), io);
        } else if (v->isBoolElementAtIndex(0)) {
          addBinaryAttributeAsSubtypeToDataSet(*i, "bool vect:" + ss.str(), chaos::DataType::SUB_TYPE_BOOLEAN, siz, io);
        } else if (v->isInt64ElementAtIndex(0)) {
          addBinaryAttributeAsSubtypeToDataSet(*i, "int64 vect:" + ss.str(), chaos::DataType::SUB_TYPE_INT64, siz * sizeof(int64_t), io);
        } else if (v->isDoubleElementAtIndex(0)) {
          addBinaryAttributeAsSubtypeToDataSet(*i, "double vect:" + ss.str(), chaos::DataType::SUB_TYPE_DOUBLE, siz * sizeof(double), io);
        }
      }
    } else if (cd.isInt32Value(*i)) {
      addAttributeToDataSet(*i, "int32", chaos::DataType::TYPE_INT32, io);

    } else if (cd.isInt64Value(*i)) {
      addAttributeToDataSet(*i, "int64", chaos::DataType::TYPE_INT64, io);

    } else if (cd.isDoubleValue(*i)) {
      addAttributeToDataSet(*i, "double", chaos::DataType::TYPE_DOUBLE, io);

    } else if (cd.isStringValue(*i)) {
      addAttributeToDataSet(*i, "string", chaos::DataType::TYPE_STRING, io, cd.getStringValue(*i).size() + 1);

    } else if (cd.isBoolValue(*i)) {
      addAttributeToDataSet(*i, "bool", chaos::DataType::TYPE_BOOLEAN, io);

    } else if (cd.isBinaryValue(*i)) {
      uint32_t size;
      cd.getBinaryValue(*i, size);
      std::stringstream ss;
      ss << size;
      addBinaryAttributeAsSubtypeToDataSet(*i, "binary:" + ss.str(), chaos::DataType::SUB_TYPE_CHAR, size, io);
    }
  }
}

void AbstractControlUnit::addPublicDriverPropertyToDataset(bool addDriverHandlers) {
  for (int idx = 0;
       idx != accessor_instances.size();
       idx++) {
    chaos::common::data::CDWUniquePtr ret = accessor_instances[idx]->getDrvProperties();
    std::vector<std::string>          props;
    ret->getAllKey(props);

    for (std::vector<std::string>::iterator i = props.begin(); i != props.end(); i++) {
      if (ret->isCDataWrapperValue(*i)) {
        std::string                       desc;
        int                               count = 1;
        chaos::common::data::CDWUniquePtr cd    = ret->getCSDataValue(*i);
        if (cd->hasKey(PROPERTY_VALUE_DESC_KEY)) {
          desc = cd->getStringValue(PROPERTY_VALUE_DESC_KEY);
        }
        if (cd->hasKey(PROPERTY_VALUE_COUNT_KEY)) {
          count = cd->getInt32Value(PROPERTY_VALUE_COUNT_KEY);
        }
        if (cd->hasKey(PROPERTY_VALUE_KEY) && cd->hasKey(PROPERTY_VALUE_PUB_KEY)) {
          std::string attrname = cd->getStringValue(PROPERTY_VALUE_PUB_KEY);

          DataType::DataSetAttributeIOAttribute type = chaos::DataType::Bidirectional;
          if (cd->hasKey(PROPERTY_VALUE_IO_KEY)) {
            type = (DataType::DataSetAttributeIOAttribute)cd->getInt32Value(PROPERTY_VALUE_IO_KEY);
          }

          if (cd->hasKey(PROPERTY_VALUE_ALARM_LVL_KEY)) {
            addStateVariable(StateVariableTypeAlarmDEV, attrname, "Notify alarm on " + attrname);
          }
          chaos::DataType::DataType dstype = cd->getValueType(PROPERTY_VALUE_KEY);
          if (dstype == chaos::DataType::TYPE_BYTEARRAY) {
            unsigned int subtype = (unsigned int)cd->getBinarySubtype(PROPERTY_VALUE_KEY);
            addBinaryAttributeAsSubtypeToDataSet(attrname, desc, (DataType::BinarySubtype)subtype, count, type);
            ACULDBG_ << "ADDING " << chaos::DataType::subtypeDescriptionByIndex(subtype) << " Array (" << count << "):" << *i << " as " << attrname << " type:" << dstype << " IO:" << type;

          } else {
            ACULDBG_ << "ADDING ATTRIBUTE:" << *i << " as " << attrname << " type:" << dstype << " IO:" << type;

            addAttributeToDataSet(attrname, desc, dstype, type);
          }
          if (addDriverHandlers && ((type == chaos::DataType::Bidirectional) || (type == chaos::DataType::Input))) {
            if (dstype == chaos::DataType::TYPE_DOUBLE) {
              addHandlerOnInputAttributeName<AbstractControlUnit, double>(this, &AbstractControlUnit::setDrvProp, attrname);

            } else if (dstype == chaos::DataType::TYPE_INT32) {
              addHandlerOnInputAttributeName<AbstractControlUnit, int32_t>(this, &AbstractControlUnit::setDrvProp, attrname);

            } else if (dstype == chaos::DataType::TYPE_INT64) {
              addHandlerOnInputAttributeName<AbstractControlUnit, int64_t>(this, &AbstractControlUnit::setDrvProp, attrname);

            } else if (dstype == chaos::DataType::TYPE_BOOLEAN) {
              addHandlerOnInputAttributeName<AbstractControlUnit, bool>(this, &AbstractControlUnit::setDrvProp, attrname);

            } else if (dstype == chaos::DataType::TYPE_STRING) {
              addHandlerOnInputAttributeName<AbstractControlUnit, std::string>(this, &AbstractControlUnit::setDrvProp, attrname);
            }
          }
        }
      }
    }
  }
}
